# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


class FileWriter:
    def __init__(self, file_path):
        self._fpath = file_path
        self._f = open(self._fpath, "w")

    def close(self):
        self._f.close()

    def write(self, content):
        self._f.write(content)

    def _get_file(self):
        return self._f


class FileLineWriter(FileWriter):
    def write(self, line):
        self._f.write(line + os.linesep)


class FileArray2LineWriter(FileWriter):
    def __init__(self, file_path, delimiter=","):
        super().__init__(file_path)
        self.delimiter = delimiter

    def write(self, line_parts):
        super().write(self.delimiter.join(line_parts))



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import abc
from arjuna.tpi.engine.asserter import AsserterMixIn
from arjuna.interact.http.model.internal.helper.yaml import convert_yaml_obj_to_content
from arjuna.tpi.parser.text import Text

class TextValidator(AsserterMixIn, metaclass=abc.ABCMeta):

    def __init__(self, response):
        super().__init__()
        self.__response = response

    @property
    def response(self):
        return self.__response

    @abc.abstractmethod
    def _eval_pattern_exists(self, pattern, exists):
        pass

    @abc.abstractmethod
    def _eval_pattern_match(self, pattern, actual, expected):
        pass

    def __extract_patterns(self, patterns):
        match_dict = {}
        patterns = type(patterns) in {set, tuple, list} and patterns or {str(patterns)}
        for pattern in patterns:
            try:
                match_dict[pattern] = self.response.find(pattern)
            except re.error as e:
                raise Exception("Wrong Regex syntax: >>{}<<. Error: {}".format(pattern, str(e)))
            except Exception:
                match_dict[pattern] = None
        return match_dict

    def assert_has_patterns(self, patterns):
        match_dict = self.__extract_patterns(patterns)
        match_dict = {k:v is not None for k,v in match_dict.items()}
        for regex, exists in match_dict.items():
            self._eval_pattern_exists(regex, exists)

    def assert_match_for_patterns(self, patterns):
        match_dict = self.__extract_patterns([k for k in patterns.keys()])
        for regex, actual in match_dict.items():
            self._eval_pattern_match(regex, actual, patterns[regex])

class ExpectedTextValidator(TextValidator):

    def __init__(self, response):
        super().__init__(response)

    def _eval_pattern_exists(self, pattern, exists):
        self.asserter.assert_true(exists, f"Expected pattern >>{pattern}<< was not found.")

    def _eval_pattern_match(self, pattern, actual, expected):
        self.asserter.assert_equal(actual, expected, f"Value for pattern >>{pattern}<< does not match.")

class UnexpectedTextValidator(TextValidator):

    def __init__(self, response):
        super().__init__(response)

    def _eval_pattern_exists(self, pattern, exists):
        self.asserter.assert_false(exists, f"Unexpected pattern >>{pattern}<< was found.")

    def _eval_pattern_match(self, pattern, actual, expected):
        self.asserter.assert_equal(actual, expected, f"Value for pattern >>{pattern}<< does not match.")





# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from arjuna.interact.gui.auto.finder.element_finder import ElementFinder

class GuiWidgetContainer(metaclass=abc.ABCMeta):
    '''
        Container of GuiWidgets.

        **GuiAutomator** and **GuiElement** are current implementations of a GuiWidgetContainer.

        Arguments:
            config: **Configuration** object.
    '''

    def __init__(self, config):
        self.__config = config
        from arjuna.interact.gui.auto.condition.container_conditions import GuiGuiWidgetContainerConditions
        self.__element_finder = ElementFinder(self)
        self.__container_conditions = GuiGuiWidgetContainerConditions(self)

    @property
    def config(self):
        '''
            **Configuration** object associated with this GuiWidgetContainer.
        ''' 
        return self.__config

    @property
    def _element_finder(self):
        return self.__element_finder

    @property
    def max_wait(self):
        '''
            Value of **ArjunaOption.GUIAUTO_MAX_WAIT** for this GuiWidgetContainer.
        ''' 
        return self.config.value("guiauto.max.wait")

    def __elem_wait(self, wmd):
        return wmd.max_wait and wmd.max_wait or self.max_wait

    def _wait_until_element_absent(self, wmd):
        return self.__container_conditions.AbsenceOfElement(wmd).wait(max_wait=self.__elem_wait(wmd))

    def _wait_until_element_found(self, gui_element):
        return self.__container_conditions.PresenceOfElement(gui_element).wait(max_wait=self.__elem_wait(gui_element._wmd))

    def _wait_until_multielement_found(self, multi_guielement):
        return self.__container_conditions.PresenceOfMultiElement(multi_guielement).wait(max_wait=self.__elem_wait(multi_guielement._wmd))

    def _load_multielement(self, multi_guielement):
        locator_type, locator_value, size, dispatcher = self._wait_until_multielement_found(multi_guielement)
        if size == 0:
            raise Exception("MultiElement could not be found with any of the provided locators.")
        multi_guielement._located_with = locator_type, locator_value
        multi_guielement._dispatcher = dispatcher
        multi_guielement._size = size
        multi_guielement._load_source_parser()

    def _load_element(self, gui_element):
        locator_type, locator_value, size, dispatcher = self._wait_until_element_found(gui_element)
        gui_element._located_with = locator_type, locator_value
        gui_element._dispatcher = dispatcher
        gui_element._load_source_parser()


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.core.error import ExclusionRuleMet, NoInclusionRuleMet

from .rule import *

class Selector:

    def __init__(self):
        self.__irules = list()
        self.__erules = list()

        self.__package_erules = list()
        self.__package_irules = list()
        self.__module_erules = list()
        self.__module_irules = list()
        self.__test_erules = list()
        self.__test_irules = list()

        self.__frozen = False

    def include(self, rule_str):
        self.__irules.append(self.__build_rule(rule_str))

    def exclude(self, rule_str):
        self.__erules.append(self.__build_rule(rule_str))

    def __build_rule(self, rule_str):
        pattern = None
        rule_str = rule_str.replace(r'\!', '!') # From command line ! has to be escaped
        try:
            return BoolAttrPatternRule.from_str(rule_str)
        except RulePatternDoesNotMatchError:
            try:
                return IterablePatternRule.from_str(rule_str)
            except RulePatternDoesNotMatchError:
                try:
                    return AttrPatternRule.from_str(rule_str)
                except RulePatternDoesNotMatchError:
                    raise Exception("Rule is invalid: " + rule_str)

    @property
    def irules(self):
        return self.__irules

    @property
    def erules(self):
        return self.__erules

    @property
    def package_irules(self):
        return self.__package_irules

    @property
    def package_erules(self):
        return self.__package_erules

    @property
    def module_irules(self):
        return self.__module_irules

    @property
    def module_erules(self):
        return self.__module_erules

    @property
    def test_irules(self):
        return self.__test_irules

    @property
    def test_erules(self):
        return self.__test_erules

    def __freeze(self):
        for irule in self.irules:
            if irule.target == "package":
                self.__package_irules.append(irule)
            elif irule.target == "module":
                self.__module_irules.append(irule)
            else:
                self.__test_irules.append(irule)

        for erule in self.erules:
            if erule.target == "package":
                self.__package_erules.append(erule)
            elif erule.target == "module":
                self.__module_erules.append(erule)
            else:
                self.__test_erules.append(erule)

    def __validate_package_rules(self, obj):
        for rule in self.__package_erules:
            if rule.matches(obj):
                raise ExclusionRuleMet(rule)

        if not self.__package_irules:
            return

        for rule in self.__package_irules:
            if not rule.matches(obj):
                continue
            else:
                return

        raise NoInclusionRuleMet()

    def __validate_module_rules(self, obj):
        for rule in self.__module_erules:
            if rule.matches(obj):
                raise ExclusionRuleMet(rule)

        if not self.__module_irules:
            return

        for rule in self.__module_irules:
            if not rule.matches(obj):
                continue
            else:
                return

        raise NoInclusionRuleMet()

    def __validate_test_rules(self, obj):
        for rule in self.__test_erules:
            if rule.matches(obj):
                raise ExclusionRuleMet(rule)

        if not self.__test_irules:
            return

        for rule in self.__test_irules:
            if not rule.matches(obj):
                continue
            else:
                return

        raise NoInclusionRuleMet()

    def validate(self, obj):
        '''
        Rules are segregated as package, module and test rules (inclusion/exclusion).

        Following is the test selection process as per Arjuna rules:

            #. Package check: Specified using ip/ep or ir/er with "package operator operand" grammar.
                - if package for a test meets an exclusion rule, it is excluded.
                - if no inclusion rule is specified, it is included for module validation.
                - if an inclusion rule is met, it is selected for module validation.
                - if no inclusion rule is met, it is excluded.
            #. Module check: Specified using im/em or ir/er with "module operator operand" grammar.
                - if module for a test meets an exclusion rule, it is excluded.
                - if no inclusion rule is specified, it is included for test validation.
                - if an inclusion rule is met, it is selected for test validation.
                - if no inclusion rule is met, it is excluded.
            #. Test check: Specified using it/et or ir/er with any rule grammar except "package operator operand" and "module operator operand".
                - if a test meets an exclusion rule, it is excluded.
                - if no inclusion rule is specified, it is included in test group run.
                - if an inclusion rule is met, it is included in test group run.
                - if no inclusion rule is met, it is excluded from test group run.
        '''

        if not self.__frozen:
            self.__freeze()
            self.__frozen = True

        try:
            self.__validate_package_rules(obj)
        except (ExclusionRuleMet, NoInclusionRuleMet) as e:
            raise e
        else:
            try:
                self.__validate_module_rules(obj)
            except (ExclusionRuleMet, NoInclusionRuleMet) as e:
                raise e    
            else:
                self.__validate_test_rules(obj)

    def __str__(self):
        return str([str(r) for r in self.irules] + [str(r) for r in self.erules])



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from arjuna.tpi.error import *
from arjuna.core.error import *
from arjuna.tpi.tracker import track
from arjuna.tpi.guiauto.meta.locator import GuiWidgetDefinition

@track("trace")
class GuiEmdFinder:

    def __init__(self, gui):
        self.__gui = gui
        self.__automator = gui._automator

    def __getattr__(self, name):
        return partial(getattr(self.__automator, name.lower()), self.__gui)

@track("trace")
class GuiFinder:

    def __init__(self, gui):
        self.__wmd_finder = GuiEmdFinder(gui)
        self.__gui = gui

    def __process_widgetdef_in_relations(self, wmd):
        from arjuna import log_debug        
        from arjuna.core.error import ArjunaTimeoutError
        from arjuna.tpi.error import GuiWidgetNotPresentError
        log_debug("Processing GuiWidgetDefinition relationships for GuiWidgetMetaData: {} in gui: {}".format(wmd, self.__gui))
        for k,v in wmd.meta.relations.items():
            from arjuna import GuiWidgetDefinition
            if isinstance(v, GuiWidgetDefinition):
                log_debug("Triggering locating operation for GuiWidgetDefinition {} in relations dict.".format(v))
                rwmd = v._as_wmd()
                rwmd = self.__process_widgetdef_in_relations(rwmd)
                log_debug("Finding element with wmd: {}.".format(rwmd))
                try:
                    wmd.meta.relations[k] = getattr(self.__wmd_finder, rwmd.meta["type"].name.lower())(rwmd).dispatcher.driver_element
                except ArjunaTimeoutError:
                    raise GuiWidgetNotPresentError(self.__gui, rwmd) 
                #  = self.locate(v).dispatcher.driver_element
                log_debug("Replaced relation {} with corresponding GuiElement in relations dict.".format(k))               
        return wmd

    def locate(self, locator):
        from arjuna import log_debug
        from arjuna.core.error import ArjunaTimeoutError
        from arjuna.tpi.error import GuiWidgetNotPresentError
        
        wmd = locator._as_wmd()
        wmd = self.__process_widgetdef_in_relations(wmd)
        log_debug("Finding element with wmd: {}.".format(wmd))
        try:
            return getattr(self.__wmd_finder, wmd.meta["type"].name.lower())(wmd)
        except ArjunaTimeoutError:
            raise GuiWidgetNotPresentError(self.__gui, wmd)         

    def __locate_interim(self, name):
        def finder(fargs=None, **kwargs):
            locator = GuiWidgetDefinition(type=name, fmt_args=fargs, **kwargs)
            return self.locate(locator)
        return finder

    def __getattr__(self, name):
        return self.__locate_interim(name)

@track("trace")
class GuiElementEmdFinder:

    def __init__(self, gui_element):
        self.__gui_element = gui_element

    def element(self, wmd):
        from arjuna.tpi.guiauto.widget.element import GuiElement
        gui_element = GuiElement(self.__gui_element.gui, wmd)
        self.__gui_element._load_element(gui_element)
        return gui_element

    def multi_element(self, wmd):
        from arjuna.tpi.guiauto.widget.multielement import GuiMultiElement
        m_guielement = GuiMultiElement(self.__gui_element.gui, wmd)
        self.__gui_element._load_multielement(m_guielement)
        return m_guielement

@track("trace")
class GuiElementFinder:

    def __init__(self, gui_element):
        self.__wmd_finder = GuiElementEmdFinder(gui_element)
        self.__gui_element = gui_element

    def locate(self, locator):
        from arjuna import log_debug
        from arjuna.core.error import ArjunaTimeoutError
        from arjuna.tpi.error import GuiWidgetNotPresentError

        wmd = locator._as_wmd()
        log_debug("Finding element with wmd: {}.".format(wmd))
        try:
            return getattr(self.__wmd_finder, wmd.meta["type"].name.lower())(wmd)
        except ArjunaTimeoutError:
            raise GuiWidgetNotPresentError(self.__gui_element.gui, self.__gui_element)          

    def __locate_interim(self, name):
        def finder(fargs=None, **kwargs):
            locator = GuiWidgetDefinition(type=name, fmt_args=None, **kwargs)
            return self.locate(locator)
        return finder

    def __getattr__(self, name):
        return self.__locate_interim(name)


import copy
import re
import pickle
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.configure.validator import Validator
from arjuna.core.fmt import arj_format_str

class WithX:

    def __init__(self, xdict={}):
        def process_value(wtype, wvalue):
            if wtype in {'ATTR', 'FATTR', 'BATTR', 'EATTR'}: #:, 'NODE', 'BNODE', 'FNODE'}:
                if len(wvalue) > 1:
                    raise Exception("attr/fattr/battr/eattr specification should contain only a single key value pair for attribute name and value. Wrong withx definition found wtype: {} with value {}".format(wtype, wvalue))
                name = list(wvalue.keys())[0]
                value = list(wvalue.values())[0]
                return {'name' : name, 'value' : value}
            else:
                return wvalue

        self.__xdict = CIStringDict()
        for k,v in xdict.items():
            try:
                wname = Validator.name(k)
                wtype = xdict[k]["wtype"].strip().upper()
                wvalue = xdict[k]["wvalue"]
                self.__xdict[wname] = {"wtype" : wtype, "wvalue" : process_value(wtype, wvalue)}
            except Exception as e:
                raise Exception(f"Invalid WithX entry for name >>{k}<<. {e}")

    def has_locator(self, name):
        return name in self.__xdict

    def format_args(self, name, vargs, kwargs):
        if not self.has_locator(name):
            raise Exception("No WithX locator with name {} found.".format(name))

        # Critical to create a copy
        fmt = copy.deepcopy(self.__xdict[name])
        if kwargs:
            repl_dict = {k.lower():v for k,v in kwargs.items()}
        else:
            repl_dict = {}
        try:

            return fmt["wtype"], arj_format_str(fmt["wvalue"], vargs, repl_dict)
            # if fmt["wtype"] in {'ATTR', 'FATTR', 'BATTR', 'EATTR', 'NODE', 'BNODE', 'FNODE'}:
            #     out = dict()
            #     for k,v in fmt["wvalue"].items():
            #         if type(v) in {list, tuple}:
            #             out[_format(k, repl_dict)] = [_format(v_entry, repl_dict) for v_entry in v]
            #         else:
            #             out[_format(k, repl_dict)] = _format(v, repl_dict)
            #         #out[k.format(**kwargs)] = v.format(**kwargs)
            #     return fmt["wtype"], out
            # else:
            #     return fmt["wtype"], _format(["wvalue"], repl_dict) #fmt["wvalue"].format(*vargs, **kwargs)
        except Exception as e:
            from arjuna import log_error
            log_error(f"Error in processing withx {name} : {fmt} for vargs {vargs} and kwargs {kwargs}")
            raise

    def format(self, name, loc_obj):
        vargs = None
        kwargs = None
        if type(loc_obj) is str:
            vargs = [loc_obj]
            kwargs = dict()
        elif type(loc_obj) in {list, tuple}:
            vargs = loc_obj
            kwargs = dict()
        elif type(loc_obj) is dict:
            vargs = []
            kwargs = loc_obj
        return self.format_args(name, vargs, kwargs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import abc
from urllib.parse import urlparse, urlencode, parse_qs, quote
from requests import Request
from arjuna.tpi.error import HttpRequestCreationError
from arjuna.tpi.parser.json import Json, JsonDict, JsonList
from arjuna.tpi.parser.html import Html
from arjuna.tpi.data.entity import _DataEntity
from arjuna.tpi.engine.asserter import AsserterMixIn
import time

from .packet import HttpPacket
from .response import HttpResponse

class HttpRequest(HttpPacket):
    '''
        Encapsulates HTTP request packet.

        Arguments:
            session: `HttpSession` object which created this `HttpRequest`.
            request: `requests` library's Request object wrapped by this class.

        Keyword Arguments:
            label: Label for this request. If available, it is used in Reports and logs.
            xcodes: Expected Status Code(s)
            allow_redirects: If True, redirections are allowed for the HTTP message. Default is True.
            timeout: How long to wait for the server to send data before giving up.
            cookies: Cookie dictionary
    '''

    def __init__(self, session, request, label=None, xcodes=None, allow_redirects=True, timeout=None, cookies=None):
        super().__init__(request)
        self.__session = session
        self.__request = request
        req_repr = "{} {}".format(self.method, self.url)
        self.__label = label and label or req_repr
        printable_label = label and self.__label + "::" + req_repr or req_repr
        printable_label = len(printable_label) > 119 and printable_label[:125] + "<SNIP>" or printable_label
        self.__label = printable_label
        self.__allow_redirects = allow_redirects
        self.__timeout = timeout
        self.__cookies = cookies

    @property
    def _request(self):
        return self.__request

    @property
    def allow_redirects(self):
        '''
        True if redirects are allowed for this request.
        '''
        return self.__allow_redirects

    @property
    def timeout(self):
        '''
        Timeout for this request.
        '''
        return self.__timeout

    @property
    def label(self) -> str:
        '''
            Label for this request object.
        '''
        return self.__label

    @property
    def query_params(self) -> dict:
        '''
            URL Query Parameters for this request object.
        '''
        return parse_qs(self.url)

    @property
    def url(self) -> str:
        '''
            URL correspnding to this request message.
        '''
        return self.__request.url

    @property
    def method(self) -> str:
        '''
            HTTP Method/Verb used by this request.
        '''
        return self.__request.method

    @property
    def xcodes(self) -> set:
        '''
        Expected status codes for this request.
        '''
        return self.__xcodes

    @property
    def text(self):
        '''
            Content of this request message as Text object.
        '''
        from arjuna.tpi.parser.text import Text
        return Text(self.content)

    @property
    def content(self):
        '''
            Content of this request message as plain unformatted text.
        '''
        return self.__request.body

    __OUT ='''{}
{}{}'''

    @classmethod
    def repr_as_str(cls, *, method, url, headers, content=None):
        if content:
            content = '\n\n{}\n'.format(content)
        else:
            content = ""
        return cls.__OUT.format(
            method + ' ' + url,
            '\n'.join('{}: {}'.format(k, v) for k, v in headers.items()),
            content
        ).strip()

    def __str__(self):
        content = self.content
        if isinstance(content, bytes):
            content = content.decode()
        return self.repr_as_str(
            method = self.__request.method,
            url = self.url,
            headers = self.__request.headers,
            content = content
        )


class _HttpRequest(HttpRequest):

    def __init__(self, session, url, method, label=None, content=None, xcodes=None, headers=None, cookies=None, allow_redirects=True, auth=None, timeout=None, pretty_url=False, query_params=None, **named_query_params):
        self.__session = session
        self.__method = method.upper()
        self.__url = url
        self.__content = None
        if content is not None:
            if type(content) in {str, dict, list, tuple} or isinstance(content, JsonDict) or isinstance(content, JsonList) or isinstance(content, _DataEntity):
                self.__content = self.__session.request_content_handler(content)
            else:
                self.__content = content
        self.__xcodes = None
        if xcodes is not None:
            self.__xcodes = self._process_codes(xcodes)
        self.__query_params = query_params
        if self.__query_params is None:
            self.__query_params = dict()
        self.__query_params.update(named_query_params)
        self.__pretty_url = pretty_url
        self.__headers = {}
        self.__headers.update(session.headers)
        if headers:
            self.__headers.update(headers)
        self.__cookies = dict()
        self.__cookies.update(self.__session.cookies)
        if cookies is not None:
            self.__cookies.update(cookies)
        self.__auth = auth is not None and auth or self.__session.auth
        self.__prepare_headers()
        self.__req = self.__build_request()
        super().__init__(self.__session, self.__req, label=label, xcodes=self.__xcodes, allow_redirects=allow_redirects, timeout=timeout, cookies=self.__cookies)

    @classmethod
    def _process_codes(cls, codes):
        return type(codes) in {set, list, tuple} and codes or {int(codes)}

    def __prepare_headers(self):
        if self.__method in {'POST', 'PUT', 'PATCH', 'OPTIONS'}:
            self.__headers['Content-Type'] = self.__content.type
        if self.__method in {'GET', 'HEAD', 'DELETE'}:
            if 'Content-Type' in self.__headers:
                del self.__headers['Content-Type']

    def __build_request(self):
        parsed_uri = urlparse(self.__url)
        #self.__headers['Host'] = parsed_uri.netloc
        if self.__pretty_url:
            query = "/".join([quote(f"{k}/{v}") for k,v in self.__query_params.items()])
            url = self.__url + query
            query_params = None
        else:
            url = self.__url
            query_params = self.__query_params

        try:
            if self.__content:
                data = self.__content.content
            else:
                data = None
            req = Request(self.__method, url, data=data, headers=self.__headers, params=query_params, cookies=self.__cookies, auth=self.__auth)
            return req.prepare()
        except Exception as e:
            raise HttpRequestCreationError("{} : {}".format(e.__class__.__name__, str(e)))




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def problem_in(*ids):
    '''
        Exclusion Relation: Problem based test function exclusion with exclude_if.

        Used to create a dependency of one test function on one or more other test functions.

        Arguments:
            *ids: One or more test function ids.

        Note:
            The depended-on test functions must belong to the same module as the dependent test function.
    '''
    if not ids:
        raise Exception("Atleast one test function id must be provided in problem_in call.")
    def call():
        return ids
    return call


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import traceback

from arjuna.tpi.error import *
from arjuna.core.error import *

# This code is inspired by fluent wait concept in Selenium Webdriver
# Reference code: https://github.com/browserstack/selenium-webdriver-python/edit/master/selenium/webdriver/support/wait.py
# However, rather than a Selnium WebDriver specific and driver level concept,
# Setu aims to make it a more generic concept available beyond WebDriver and GUI automation.
# Setu makes waiting accessible through a condition rather than the other way round.
# i.e. object.condition(args).wait(), rather than wait(driver).condition(args)

class ConditionException(Exception):

    def __init__(self, msg):
        super().__init__(msg)

class Condition:

    def __init__(self, dynamic_caller, *args, **kwargs):
        self.__dynamic_caller = dynamic_caller
        self.__args = args
        self.__kwargs = kwargs
        self.__call_result = None

    def get_call_result(self):
        return self.__call_result

    def wait(self, *, max_wait=60, poll_interval=0.5):
        from arjuna import log_debug
        log_debug("Dynamic wait call triggered with max_wait: {}".format(max_wait))
        end_time = time.time() + max_wait
        e = None
        etrace = None
        while(True):
            try:
                if self.is_met():
                    return self.get_call_result()
            except ConditionException as ce:
                raise ce
            except WaitableError as we:
                e = we
                import traceback
                etrace = traceback.format_exc()
                pass
            except Exception as f:
                import traceback
                raise Exception("An unexpected exception occured in dynamic wait. Message: {}. Trace: {}".format(str(f), traceback.format_exc()))
            time.sleep(poll_interval)
            ctime = time.time()
            if(ctime > end_time):
                break
        raise ArjunaTimeoutError(self.__class__.__name__, str(e) + etrace)

    def execute(self):
        try:
            self.__call_result = self.__dynamic_caller.call(*self.__args, **self.__kwargs)
        except ArjunaTimeoutError as e:
            raise ArjunaTimeoutError("ArjunaTimeOut: {}".format(str(e)))

class CommandCondition(Condition):

    def __init__(self, dynamic_caller, returns_none=True):
        super().__init__(dynamic_caller)
        self.__returns_none = returns_none

    def is_met(self):
        self.execute()
        if self.__returns_none: return True
        return self.get_call_result() is not None

class BooleanCondition(Condition):

    def __init__(self, dynamic_caller, expectation_type=True):
        super().__init__(dynamic_caller)
        self.__expectation_type = expectation_type

    def is_met(self):
        try:
            self.execute()
            return self.__expectation_type == self.get_call_result()
        except Exception as e:
            raise ConditionException("Unexpected exception in boolean condition checking: " + str(e) + traceback.format_exc())


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time

def main(*args, ext_engine=False):
    print("Executing Arjuna command line with args: {}".format(args))
    try:
        import signal
        import sys
        def signal_handler(sig, frame):
                print('Exiting...')
                sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler)
        from arjuna import _ArjunFacade
        facade = _ArjunFacade()
        # For external test engine, Arjuna is loaded but it does not itself execute tests.
        if ext_engine:
            facade.load(args and args or sys.argv)
        else:
            facade.launch(args and args or sys.argv)
    except Exception as e:
        # The following sleep is to accommodate a common IDE issue of
        # interspersing main exception with console output.
        time.sleep(0.5)
        msg = '''
    {0}
    Sorry. Looks like this is an error Arjuna couldn't handle.
    Create a bug report on GitHub: https://github.com/rahul-verma/arjuna
    {0}

    Message: {1}
        '''.format("-" * 70, str(e))
        print(msg)
        import traceback
        print(traceback.format_exc())



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Arjuna Defined Types

These types are used for annotating functions, methods and return types.

In future, they could be used to enforce types. For the time being, the purpose is understandability and documentation.
'''
from typing import TypeVar
from arjuna.tpi.constant import ArjunaOption

ListOrTuple = TypeVar('ListOrTuple', list, tuple)
ListOrTupleOrStr = TypeVar('ListOrTupleOrStr', list, tuple, str)
ArjunaOptionOrStr = TypeVar('ArjunaOptionOrStr', ArjunaOption, str)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna import Arjuna
from arjuna.interact.gui.auto.finder.wmd import SimpleGuiWidgetMetaData
from arjuna.tpi.guiauto.source.frame import _GuiFrameSource

from arjuna.core.poller.conditions import *
from arjuna.core.poller.caller import *
from arjuna.core.error import ChildWindowNotFoundError

class FrameConditions:

    def __init__(self, frame):
        self.__frame = frame

    @property
    def frame(self):
        return self.__frame

    def FrameIsPresent(self, wmd, *args, **kwargs):
        caller = DynamicCaller(self.frame._find_frame, wmd, *args, **kwargs)
        return CommandCondition(caller)

class FrameContainer:

    def __init__(self, gui):
        self.__gui = gui
        self.__automator = gui._automator
        self.__conditions = FrameConditions(self)

    @property
    def max_wait(self):
        return self.__automator.config.guiauto_max_wait

    def __check_tag(self, wrapped_element):
        tag = wrapped_element.source.tag
        if tag.lower() != "iframe":
            raise Exception("The element should have a 'iframe' tag for IFrame element. Found: " + tag)

    def frame(self, wmd):
        return self.__conditions.FrameIsPresent(wmd).wait(max_wait=self.max_wait)

    def _find_frame(self, wmd):
        found = False
        frame = None
        for locator in wmd.locators: 
            try:
                if locator.ltype.name == "INDEX":
                    index = locator.lvalue
                    wmd = SimpleGuiWidgetMetaData("xpath", "//iframe")
                    multi_element = self.__automator.multi_element(self.gui, wmd)
                    # multi_element.find()
                    try:
                        wrapped_element = multi_element[index]
                    except:
                        # In case another identifier is present it should be tried.
                        continue
                    self.__check_tag(wrapped_element)
                    frame = IPartialFrame(self.gui, self, multi_element, wrapped_element)
                else:
                    wmd = SimpleGuiWidgetMetaData(locator.ltype.name, locator.lvalue)
                    wrapped_element = self.__automator.element(self.gui, wmd)
                    # wrapped_element.find()
                    self.__check_tag(wrapped_element)
                    frame = IFrame(self.gui, self, wrapped_element)

                return frame
            except WaitableError as f:
                continue  
            except Exception as e:
                raise Exception(str(e) + traceback.format_exc()) 

        raise ChildFrameNotFoundError(*wmd.locators)

    def enumerate_frames(self):
        self.focus()
        wmd = SimpleGuiWidgetMetaData("xpath", "//iframe")
        multi_element = self.__automator.create_multielement(wmd)
        ret_str = os.linesep.join(["--> " + s for s in multi_element.source._get_root_content_as_list()])
        return self._source_parser.get_root_content() + os.linesep + ret_str

class DomRoot(FrameContainer):

    def __init__(self, gui):
        super().__init__(gui)
        self.__frame_context = "root"
        self._source_parser = self.__automator.source

    @property
    def frame_context(self):
        return self.__frame_context

    def __set_frame_context_str(self, name):
        self.__frame_context = name
        Arjuna.get_logger().debug("Automator is in {} frame".format(self.__frame_context))
        print("Automator is in {} frame".format(self.__frame_context))

    def set_frame_context(self, frame):
        self.__set_frame_context_str(frame)

    def set_frame_context_as_root(self):
        self.__set_frame_context_str("root")  

    def is_in_root_context(self):
        return self.__frame_context == "root"

    def focus(self):
        self.__automator.dispatcher.focus_on_dom_root()
        self.set_frame_context_as_root() 

    @property
    def source(self):
        self.focus()
        return self.__automator.source

class IFrame(FrameContainer):

    def __init__(self, gui, dom_root, wrapped_element):
        super().__init__(gui)
        self.__dom_root = dom_root
        self.__parent_frames = []
        self.__wrapped_element = wrapped_element
        self._source_parser = _GuiFrameSource(self)

    @property
    def dom_root(self):
        return self.__dom_root

    @property
    def wrapped_element(self):
        return self.__wrapped_element

    def set_parents(self, parents):
        self.__parent_frames = parents

    def _act(self, json_dict):
        return self.dom_root._act(json_dict)

    def _focus_on_parents(self):
        if self.__parent_frames:
            for parent in self.__parent_frames:
                parent.focus()

    def __reload_wrapped_element(self):
        self.wrapped_element.find()

    def focus(self):
        if not self.dom_root.is_in_root_context():
            self.dom_root.focus()
            self._focus_on_parents()
        # self.wrapped_element.find()
        self._source_parser.set_root_source(self.wrapped_element.source.content.root)
        self.__automator.dispatcher.focus_on_frame(self.wrapped_element.dispatcher)
        self.dom_root.set_frame_context(self)

    def _get_html_content_from_remote(self):
        return self.__automator.source

    def get_wrapped_element(self):
        return self.wrapped_element

    @property
    def source(self) -> _GuiFrameSource:
        self._source_parser._load()
        return self._source_parser

    # def focus_on_parent(self):
    #     self._act(TestAutomatorActionBodyCreator.jump_to_parent_frame())
    #     if self.__parent_frames:
    #         self.dom_root.set_frame_context(self.__parent_frames[-1])
    #     else:
    #         self.dom_root.set_frame_context_as_root()

    @property
    def parent(self):
        if self.__parent_frames:
            return self.__parent_frames[-1]
        else:
            return self.dom_root


class IPartialFrame(IFrame):

    def __init__(self, gui, dom_root, melement, wrapped_element):
        super().__init__(gui, dom_root, wrapped_element)
        self.__melement = melement

    def focus(self):
        # self.__melement.find()
        self.__automator.dispatcher.focus_on_frame(self.wrapped_element.dispatcher)
        self.dom_root.set_frame_context(self)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Arjuna Date Time Helper Classes

Contains many general purpose type abstractions.
'''


from arjuna.tpi.constant import TimeUnit
from datetime import timedelta, datetime

_DEF_FORMAT = '%d.%m.%y %H:%M:%S'

class Time:
    '''
        A time object.

        Arguments:
            time_unit: TimeUnit enum value.
            value: An integer representing this time value.
    '''

    def __init__(self, time_unit, value):
        self.__unit = time_unit
        self.__value = value

    @classmethod
    def seconds(cls, value):
        '''
            Factory method to a create a Time object for seconds.

            Arguments:
                value: Number of seconds
        '''
        return Time(TimeUnit.SECONDS, value)

    @classmethod
    def milliseconds(cls, value):
        '''
            Factory method to a create a Time object for milliseconds.

            Arguments:
                value: Number of milliseconds
        '''
        return Time(TimeUnit.MILLI_SECONDS, value)

    @classmethod
    def minutes(cls, value):
        '''
            Factory method to a create a Time object for minutes.

            Arguments:
                value: Number of minutes
        '''
        return Time(TimeUnit.MINUTES, value)


class DateTimeStepper:
    '''
        An iterable object that returns DateTime objects or strings as per provided arguments.

        Keyword Arguments:
            start: Arjuna DateTime object to start stepping. If not supplied current date & time is used. Returned as first value.
            delta: A DateTimeDelta object used to create next date time object. If not provided, a delta of 1 second is used.
            max_steps: Maximum number of steps or iterations of this object. Default is 100000.
            forward: Boolen value that sets the direction of date time. If True steps are directed towards future else towards past. Default is True.
            format: String format to represent the generated date time object. If not provided, it is taken from start argument. If start argument is not provided it defaults to '%d.%m.%y %H:%M:%S'.
            as_str: Boolena value that controls the return type on next calls. If True, the iterable returns the data time object as a string else DateTime object is returned. Default is True.

        Note:
            You can make next() call to get next DateTime object or string:

            .. code-block:: python

                dtstepperDateTimeStepper()
                dt = dtstepper.next()

            If you have explcitly set max_steps, a better usage pattern is to use this object in a Python for loop:

            .. code-block:: python

                for dt in DateTimeStepper(max_steps=25):
                    # Do something
                    print(dt)
    '''

    def __init__(self, *, start: 'DateTime'=None, delta: 'DateTimeDelta'=None, max_steps: int=100000, forward: bool=True, format: str=None, as_str: bool=True):
        if start is not None:
            self.__start = start
            if format is not None:
                self.__format = format
            else:
                self.__format = start._format
        else:
            if format is not None:
                self.__start = DateTime(datetime.today(), format=format)
                self.__format = format
            else:
                self.__start = DateTime(datetime.today(), format=_DEF_FORMAT)
                self.__format = _DEF_FORMAT

        self.__delta = delta is not None and delta or DateTimeDeltaBuilder().seconds(1).build()
        self.__max_steps = max_steps
        self.__forward = forward
        self.__as_str = as_str
        self.__current = DateTime(self.__start._value, format=self.__format)
        self.__counter = 0

    def __iter__(self):
        return self

    def next(self):
        '''
            Returns next DateTime object or string.
        '''
        self.__counter += 1
        if self.__counter > self.__max_steps:
            raise StopIteration("Finished all steps.")
        if self.__counter != 1:
            if self.__forward:
                self.__current = DateTime((self.__current + self.__delta)._value)
            else:
                self.__current = DateTime((self.__current - self.__delta)._value)
                
        return self.__as_str and self.__current.as_str(format=self.__format) or self.__current

    __next__ = next

class DateTime:
    '''
        Represents a (mutable) date time object.

        Arguments:
            pydtobj: Python datetime object.
            format: String representation. Default is '%d.%m.%y %H:%M:%S'.

        Note:

            You can use + and - operators with DateTime and DateTimeDelta objects.

            .. code-block:: python

                dtobj + dtdelta
                dtobj - dtdelta

            Although reassignment is not necessary, you can also do

            .. code-block:: python

                dtobj = dtobj + dtdelta
                dtobj = dtobj - dtdelta
    '''

    def __init__(self, pydtobj, *, format=_DEF_FORMAT):
        self.__pydtobject = pydtobj
        self.__format = format

    @property
    def _value(self):
        return self.__pydtobject

    @property
    def _format(self):
        return self.__format

    def stepper(self, *, delta: 'DateTimeDelta'=None, max_steps: int=100000, forward: bool=True, format: str=None, as_str: bool=True):
        '''
            Factory method to create DateTimeStepper object which takes this DateTime as starting point.

            Keyword Arguments:
                delta: A DateTimeDelta object used to create next date time object. If not provided, a delta of 1 second is used.
                max_steps: Maximum number of steps or iterations of this object. Default is 100000.
                forward: Boolean value that sets the direction of date time. If True steps are directed towards future else towards past. Default is True.
                format: String format to represent the generated date time object. If not provided, it is taken from this DateTime object.
                as_str: Boolean value that controls the return type on next calls. If True, the iterable returns the data time object as a string else DateTime object is returned. Default is True.
        '''
        return DateTimeStepper(start=self, delta=delta, max_steps=max_steps, forward=forward, format=format, as_str=as_str)

    @classmethod
    def now(self, *, format: str=_DEF_FORMAT):
        '''
            Factory method to create a DateTime object representing current date and time.

            Keyword Arguments:
                format: String format to represent the generated date time object. Default is '%d.%m.%y %H:%M:%S'.
        '''
        return DateTime(datetime.today(), format=format)

    @classmethod
    def from_str(self, dtstr: str, *, format: str=_DEF_FORMAT):
        '''
            Factory method to create a DateTime object from provided date time string.

            Arguments:
                dtstr: Date Time string

            Keyword Arguments:
                format: String format to parse the provided date time string. Default is '%d.%m.%y %H:%M:%S'.
        '''
        return DateTime(datetime.strptime(dtstr, format), format=format)

    @classmethod
    def convert(self, dtstr: str, *, from_format: str=_DEF_FORMAT, to_format: str=_DEF_FORMAT):
        '''
            Factory method to create a DateTime object from provided date time string.

            Arguments:
                dtstr: Date Time string

            Keyword Arguments:
                from_format: String format to parse the provided date time string. Default is '%d.%m.%y %H:%M:%S'.
                to_format: String format to create the converted date time string. Default is '%d.%m.%y %H:%M:%S'.

            Returns:
                Date Time string as per to_format
        '''
        return DateTime.from_str(dtstr, format=from_format).as_str(format=to_format)  

    def as_str(self, *, format=None):
        '''
            String representation of this DateTime object.

            Keyword Arguments:
                format: String format to represent the provided date time object. If not provided format set during construction of this object is used.
        '''
        if format is None:
            format = self.__format
        return datetime.strftime(self.__pydtobject, format)

    def add(self, dtdelta):
        '''
            Add a DateTimeDelta object to this object. Modifies current object.

            Arguments:
                dtdelta: DateTimeDelta Object.

            Returns:
                This DateTime object (Self)
        '''
        self.__pydtobject  = self.__pydtobject + dtdelta._value
        return self

    def sub(self, dtdelta):
        '''
            Add a DateTimeDelta object to this object. Modifies current object.

            Arguments:
                dtdelta: DateTimeDelta Object.

            Returns:
                This DateTime object (Self)
        '''
        self.__pydtobject  = self.__pydtobject - dtdelta._value
        return self

    __add__ = add
    __sub__ = sub


class DateTimeDeltaBuilder:
    '''
        Builder class to flexibly build a DateTimeDelta object.
    '''

    def __init__(self):
        self.__dtdelta_kwargs = dict()

    def weeks(self, count: int):
        '''
            Set number of detla weeks.

            Arguments:
                count: Number of weeks
        '''
        self.__dtdelta_kwargs['weeks'] = count
        return self

    def days(self, count: int):
        '''
            Set number of detla days.

            Arguments:
                count: Number of days
        '''
        self.__dtdelta_kwargs['days'] = count
        return self

    def hours(self, count: int):
        '''
            Set number of detla hours.

            Arguments:
                count: Number of hours
        '''
        self.__dtdelta_kwargs['hours'] = count
        return self

    def minutes(self, count: int):
        '''
            Set number of detla minutes.

            Arguments:
                count: Number of minutes
        '''
        self.__dtdelta_kwargs['minutes'] = count
        return self

    def seconds(self, count: int):
        '''
            Set number of detla seconds.

            Arguments:
                count: Number of seconds
        '''
        self.__dtdelta_kwargs['seconds'] = count
        return self

    def milliseconds(self, count: int):
        '''
            Set number of detla miliseconds.

            Arguments:
                count: Number of milliseconds
        '''
        self.__dtdelta_kwargs['milliseconds'] = count
        return self

    def microseconds(self, count: int):
        '''
            Set number of detla microseconds.

            Arguments:
                count: Number of microseconds
        '''
        self.__dtdelta_kwargs['microseconds'] = count
        return self

    def build(self):
        '''
            Build a DateTimeDelta object as per values set in builder.

            Returns:
                DateTimeDelta object.
        '''
        return DateTimeDelta(**self.__dtdelta_kwargs)


class DateTimeDelta:
    '''
        Represents a delta (difference) in date time.

        Keyword Arguments:
            weeks: Delta number of weeks.
            days: Delta number of days.
            hours: Delta number of hours.
            minutes: Delta number of minutes.
            seconds: Delta number of seconds.
            miliseconds: Delta number of miliseconds.
            microseconds: Delta number of microseconds.
    '''

    def __init__(self, *, weeks: int=0, days: int=0, hours: int=0, minutes: int=0, seconds: int=0, milliseconds: int=0, microseconds: int=0):
        self.__delta = timedelta(weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds, milliseconds=milliseconds, microseconds=microseconds)

    @property
    def _value(self):
        return self.__delta

    @classmethod
    def builder(cls):
        '''
            Create a DateTimeDeltaBuilder object.

            Returns:
                DateTimeDeltaBuilder object
        '''
        return DateTimeDeltaBuilder()

    @classmethod
    def zero(cls):
        '''
            Create a DateTimeDelta object which represents no change in date time.

            Returns:
                DateTimeDelta object.
        '''
        return DateTimeDelta()

    def from_now(self, *, forward: bool=True, format: str=_DEF_FORMAT):
        '''
            Keyword Arguments:
                forward: Boolen value that sets the direction of date time. If True delta is added else it is subtracted. Default is True.
                format: String format to represent the generated date time object. Default is '%d.%m.%y %H:%M:%S'.
        '''
        if forward:
            return DateTime(datetime.today() + self._value, format=format)
        else:
            return DateTime(datetime.today() - self._value, format=format)




import sys
import os

from arjuna.main import main
args = sys.argv
project_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))
args.extend(("-p", project_dir))
main(*args)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class GuiNameStore:

    def __init__(self):
        # dict<String, GuiNameStore>
        self.__ns_map = {}

    # Needs to be thread safe
    def has_namespace(self, name):
        return name in self.__ns_map

    # loader is GuiNamespaceLoader
    # Needs to be thread safe
    def load_namespace(self, name, loader):
        if not self.has_namespace(name):
            loader.load()
            self.__ns_map[name.lower()] = loader.namespace

        return self.__ns_map[name.lower()]

    # Needs to be thread-safe
    def get_namespace(self, name):
        return self.__ns_map[name.lower()]


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from .factory import DataReference
from .reference import ContextualDataReference
from arjuna.tpi.helper.arjtype import CIStringDict

from arjuna.core.utils import file_utils
from arjuna.tpi.parser.json import Json, JsonDict

class L10NRef:

    def __init__(self, ref_config):
        self.map = {}
        self.ref_config = ref_config

    def update_from_excel_ref(self, localizer_ref):
        for lang, record in localizer_ref.map.items():
            if lang not in self.map:
                self.map[lang] = JsonDict()
            self.map[lang].update(record.named_values)

    def update_from_json_ref(self, json_ref):
        for lang, map in json_ref.map.items():
            if lang not in self.map:
                self.map[lang] = JsonDict()
            self.map[lang].update(map)

    def lang(self, lang):
        if lang.lower() in self.map:
            return self.map[lang.lower()]
        else:
            raise Exception("Language key {} not found in localizer: {}.".format(lang, self.__class__.__name__))

    def __str__(self):
        return str({k: str(v) for k,v in self.map.items()})

    def enumerate(self):
        for k,v in self.map.items():
            print(k, "::", type(v), str(v))

    def get_localizer_file_path(self, ldir, fpath):
        if file_utils.is_absolute_path(fpath):
            if not file_utils.is_file(fpath):
                if file_utils.is_dir(fpath):
                    raise Exception("Not a file: {}".format(fpath))
                else:
                    raise Exception("File does not exist: {}".format(fpath))
            return fpath
        else:
            fpath = os.path.abspath(os.path.join(ldir, fpath))
            if not file_utils.is_file(fpath):
                if file_utils.is_dir(fpath):
                    raise Exception("Not a file: {}".format(fpath))
                else:
                    raise Exception("File does not exist: {}".format(fpath))
            return fpath

class ExcelL10NRef(L10NRef):

    def __init__(self, ref_config, fpath):
        super().__init__(ref_config)
        excel_ref = DataReference.create_contextual_data_ref(self.ref_config, fpath)
        self.map = excel_ref._map

class JsonL10NRef(L10NRef):

    def __init__(self, ref_config, ldir):
        from arjuna import log_fatal
        super().__init__(ref_config)
        fnames = os.listdir(ldir)
        fnames.sort()
        for fname in fnames:
            if fname.lower().endswith(".json"):
                lang = os.path.splitext(fname)[0].replace("-","_").lower()
                fpath = self.get_localizer_file_path(ldir, fname)
                try:
                    lang_map = Json.from_file(fpath)
                except Exception as e:
                    log_fatal("Error in processing l10n file: {}".format(fpath))
                    log_fatal("Error message: {}".format(str(e)))
                    import traceback
                    log_fatal("Trace: {}".format(traceback.format_exc()))
                    raise
                self.map[lang] = lang_map

class Localizers:

    def __init__(self):
        vars(self)['_store'] = CIStringDict()

    def keys(self):
        return self._store.keys()

    def __getitem__(self, name):
        return self._store[name]

    def __setitem__(self, name, value):
        self._store[name] = value

    def __getattr__(self, name):
        if type(name) is str and not name.startswith("__"):
            return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __str__(self):
        return str(self._store)

    def has(self, name):
        return name.lower() in self._store

class Localizer:

    def __init__(self, global_map, buckets):
        self.__globals = global_map
        self.__buckets = buckets     
        self.__bucket_names = self.__buckets.keys()

    @property
    def globals(self):
        return self.__globals

    @property
    def buckets(self):
        return self.__buckets

    @property
    def bucket_names(self):
        return self.__bucket_names

    @classmethod
    def __process_json_ref(cls, l10n_merged_ref, l10n_refs, json_ref, bucket="root"):
        # Check this logic for locale conversion.
        l10n_merged_ref.update_from_json_ref(json_ref)
        if not l10n_refs.has(bucket):
            l10n_refs[bucket] = json_ref

    @classmethod
    def load_all(cls, ref_config):
        from arjuna.tpi.constant import ArjunaOption
        l10n_dir = ref_config.value(ArjunaOption.L10N_DIR)
        l10n_merged_ref = L10NRef(ref_config)
        l10n_refs = Localizers()
        if os.path.isdir(l10n_dir):
            fnames = os.listdir(l10n_dir)
            fnames.sort()
            for fname in fnames:
                if fname.lower().endswith("xls"):
                    ref = ExcelL10NRef(ref_config, os.path.join(l10n_dir, fname))
                    l10n_merged_ref.update_from_excel_ref(ref)
                    l10n_refs[os.path.splitext(fname)[0]] = ref

        if os.path.isdir(l10n_dir):        
            json_ref = JsonL10NRef(ref_config, l10n_dir)
            cls.__process_json_ref(l10n_merged_ref, l10n_refs, json_ref, bucket="root")
            buckets = os.listdir(l10n_dir)
            for bucket in buckets:
                if os.path.isdir(os.path.join(l10n_dir, bucket)):
                    dpath = os.path.join(l10n_dir, bucket)
                    json_ref = JsonL10NRef(ref_config, dpath)
                    cls.__process_json_ref(l10n_merged_ref, l10n_refs, json_ref, bucket=bucket)

        # l10n_merged_ref.enumerate()
        return Localizer(l10n_merged_ref, l10n_refs)

def L(in_str, *, locale=None, bucket=None, strict=None):
    from arjuna import Arjuna, ArjunaOption
    bucket = bucket
    query = in_str
    if bucket is None:
        if in_str.find('.') != -1:
            bucket, query = in_str.split('.', 1)
            bucket = bucket.lower()
            if bucket not in Arjuna.get_localizer().bucket_names:
                bucket = None
                query = in_str
        else:
            query = in_str
    lang = locale and locale.name.lower() or Arjuna.get_config().l10n_locale.name.lower()
    try:
        if not bucket:
            val = Arjuna.get_localizer().globals.lang(lang)[query]
        else:
            val = Arjuna.get_localizer().buckets[bucket].lang(lang)[query]
        if not val:
            raise Exception("No localized string found for: {} for Locale.{}".format(in_str, lang.upper()))
        else:
            return val
    except Exception as e:
        if strict is None:
            strict_mode = Arjuna.get_config().value(ArjunaOption.L10N_STRICT)
        else:
            strict_mode = strict

        if strict_mode:
            import traceback
            raise Exception("Error in retrieving localized string for: {}. {}. {}".format(in_str, str(e), traceback.format_exc()))
        else:
            return in_str



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Classes to assist in XML Parsing. 
'''
import os
import re
from lxml import etree, html
from typing import List, Dict, Tuple

from arjuna.tpi.tracker import track
from arjuna.tpi.helper.arjtype import CIStringDict

def _process_tags(tagsvalue):
    tag_list = None
    if type(tagsvalue) is str:
        tag_list = tagsvalue.strip().split()
    else:
        tag_list = tagsvalue
    return [t.lower()=='any' and '*' or t for t in tag_list]

@track("trace")
class NodeLocator:
    '''
        Locator for finding an XML Node in an **XmlNode**.

        Keyword Arguments:
            tags: (Optional) Descendant tags for the node. Can be a string of single or multiple tags or a list/tuple of tags.
            **attrs: Arbitrary number of key value pairs representing attribute name and value.

        Raises:
            Exception: If neither tag nor an attribute is provided.

        Note:
            You can use tag and attributes in combination.

            Supports nested node finding.
    '''
    
    def __init__(self, *, tags: 'strOrSequence'=None, **attrs):

        if tags is None and not attrs:
            raise Exception("You must provided tags and/or attributes for finding nodes.")

        attr_conditions = []
        if attrs:
            for attr, value in attrs.items():
                if value is None:
                    attr_conditions.append("@{}".format(attr))
                else:
                    attr_conditions.append("contains(@{}, '{}')".format(attr, value))

        attr_str = ""
        if attr_conditions:
            attr_str = "[{}]".format("and ".join(attr_conditions))
        tags = tags and "//".join(_process_tags(tags)) or "*"
        prefix = ".//"

        self.__xpath = "{}{}{}".format(prefix, tags, attr_str)

    def search_node(self, node: 'XmlNode') -> tuple:
        '''
        Search `XmlNode` objects that match this locator in the provided `XmlNode` object.
        '''
        return (XmlNode(n) for n in node.xpath(self.__xpath))


def _process_child_html(in_str):
    processed = "\n".join([l for l in in_str.splitlines() if l.strip()])
    return "\t" + processed

def _remove_empty_lines_from_string(in_str):
    return '\n'.join(
        [l.strip() for l in in_str.splitlines() if l.strip()]
    )

def _empty_or_none(in_str):
    if type(in_str) is str and not in_str.strip():
        return True
    else:
        return in_str is None


@track("trace")
class XmlNode:
    '''
        Represents a single node in a parsed XML.

        Arguments:
            node: **lxml** Element object.
    '''

    def __init__(self, node):
        self.__node = node
        self.__attrs = CIStringDict(self.node.attrib)

    @property
    def node(self):
        '''
            Wrapped **lxml** Element

            Not supposed to be used directly.
        '''
        return self.__node

    def get_text(self, normalize: bool=False) -> str:
        '''
            Text of this node.

            Keyword Arguments:
                normalize: If True, empty lines are removed and individual lines are trimmed.
        '''
        texts = self.texts

        if normalize:
            return "".join([l for l in texts if l !="\n"]).strip()
        else:
            return "".join(texts).strip()

    @property
    def normalized_text(self) -> str:
        '''
            Text of this node with empty lines removed and individual lines trimmed.
        '''
        return self.get_text(normalize=True)

    @property
    def text(self) -> str:
        '''
            Unaltered text of the node.
        '''
        text = self.get_text()
        if text is None:
            return ""
        else:
            return text

    @property
    def texts(self) -> list:
        '''
            List of Texts of the node.

            Note:
                Multiple texts are stored separately.
        '''
        return self.node.xpath(".//text()")

    def get_inner_xml(self, normalize=False) -> str:
        '''
            Inner XML of this node.

            Keyword Arguments:
                normalize: If True, empty lines are removed between children nodes.
        '''

        def same(i):
            return i

        processor = normalize and _process_child_html or same
        out = [
                processor(etree.tostring(c, encoding='unicode'))
                for c in list(self.__node.iterchildren())
            ]
        return "\n".join(out).strip()

    @property
    def inner_xml(self) -> str:
        '''
            Unaltered inner XML of this node
        '''
        return self.get_inner_xml()

    @property
    def normalized_inner_xml(self) -> str:
        '''
            Normalized inner XML of this node, with empty lines removed between children nodes.
        '''
        return self.get_inner_xml(normalize=True)

    def remove_all_children(self) -> None:
        '''
            Remove all children nodes from this node.
        '''
        for child in list(self.__node): self.__node.remove(child)

    def as_str(self, normalize=False) -> str:
        '''
            String representation of this node.

            normalize: If True all new lines are removed and more than one conseuctive space is converted to a single space.
        '''
        true_source = etree.tostring(self.node, encoding='unicode')
        if not normalize:
            return true_source
        else:
            ret_source = ' '.join(true_source.splitlines())
            return re.sub(r"\s+", " ", ret_source)

    @property
    def source(self) -> str:
        '''
            Unalereted string representation of this node.
        '''
        return self.as_str()

    @property
    def normalized_source(self) -> str:
        '''
            String representation of this node with all new lines removed and more than one conseuctive space converted to a single space.
        '''
        return self.as_str(normalize=True)

    @property
    def tag(self) -> str:
        '''
            Tag of the node.
        '''     
        return self.node.tag

    @property
    def children(self) -> Tuple['XmlNode']:
        '''
            All Children of this node as a Tuple of XmlNodes
        '''
        return (XmlNode(c) for c in list(self.node))

    @property
    def parent(self) -> 'XmlNode':
        '''
            Parent XmlNode
        '''
        return XmlNode(self.node.getparent())

    @property
    def preceding_sibling(self) -> 'XmlNode':
        '''
            The XmlNode before this node at same hierarchial level.
        '''
        return XmlNode(self.node.getprevious())

    @property
    def following_sibling(self) -> 'XmlNode':
        '''
            The XmlNode after this node at same hierarchial level.
        '''
        return XmlNode(self.node.getnext())

    @property
    def attrs(self) -> CIStringDict:
        '''
            All Attributes of this node as a dictionary.
        '''
        return self.__attrs

    def attr(self, name) -> str:
        '''
            Value of an attribute of this node.
        '''
        return self.__attrs[name]

    @property
    def value(self) -> str:
        '''
            Value of an 'value' attribute of this node.
        '''
        return self.attr("value")

    def has_attr(self, name):
        '''
            Check if an attribute is present.
        '''
        return name in self.__attrs

    def __xpath(self, xpath):
        if not xpath.startswith("."):
            return "." + xpath
        else:
            return xpath

    def findall_with_xpath(self, xpath) -> List['XmlNode']:
        '''
            Find all XmlNodes that match an XPath.
        '''
        return [XmlNode(n) for n in self.node.xpath(self.__xpath(xpath))]

    def find_with_xpath(self, xpath, position=1):
        '''
            Find nth XmlNode that matches an XPath.

            Args:
                xpath: XPath string
                position: XPath index. Default is 1.
        '''
        try:
            all = self.findall_with_xpath(xpath)
            return all[position-1]
        except IndexError as e:
            raise Exception(f"No node match at position >>{position}<< for xpath >>{xpath}<< in xml >>{self}<<")

    def findall(self, *node_locators, stop_when_matched: bool=False) -> List['XmlNode']:
        '''
            Find all XmlNodes that match one of more `NodeLocator` s.

            Args:
                *node_locators: One or more `NodeLocator` s

            Keyword Arguments:
                stop_when_matched: If True, the call returns nodes found by the first `NodeLocator` that locates one or more nodes. Default is False.

            Returns:
                List of `XmlNode` s. In case of no match, empty list is returned.
        '''

        out = []
        for locator in node_locators:
            try:
                nodes = list(locator.search_node(self.node))
            except:
                continue
            else:
                out.extend(nodes)
                if stop_when_matched:
                    if nodes:
                        break
        return out

    def find(self, *node_locators, strict: bool=False) -> 'XmlNode':
        '''
            Find first `XmlNode` that match one of more `NodeLocator` s.

            Args:
                *node_locators: One or more `NodeLocator` s

            Keyword Arguments:
                strict: If True, the call raises an exception if element is not found, else returns None
        '''
        matches = self.findall(*node_locators, stop_when_matched=True)
        if matches:
            return matches[0]
        else:
            if strict:
                raise Exception("Element could not be found with Node Locators: >><<".format([str(n) for n in node_locators]))
            else:
                return None

    def find_keyvalue_texts(self, key_locator, value_locator) -> Tuple[str, str]:
        '''
            Returns texts of first XmlNodes for a pair of `NodeLocator` s

            Args:
                key_locator: `NodeLocator` (for key)
                value_locator: First `NodeLocator` (for value)

            Returns:
                2-element tuple containing the text strings.
        '''
        key = self.find(key_locator).text
        value = self.find(value_locator).text
        return key,value

    def __str__(self):
        return self.as_str()

    def clone(self) -> 'XmlNode':
        '''
            Create a clone of this XmlNode object.
        '''
        return Xml.from_str(str(self))
        
class Xml:
    '''
        Helper class to create XmlNode objects.
    '''

    @classmethod
    def from_str(cls, xml_str):
        '''
            Create an `XmlNode` from a string.
        '''
        lenient_parser = etree.XMLParser(encoding='utf-8', recover=True)
        return XmlNode(etree.parse(StringIO(xml_str), lenient_parser))


    @classmethod
    def from_file(cls, file_path: str) -> XmlNode:
        '''
            Creates an `XmlNode` from file.

            Arguments:
                file_path: Absolute path of the json file.

            Returns:
                Arjuna's `XmlNode` object
        '''

        with open(file_path, 'r') as f:
            return cls.from_str(f.read())

    @classmethod
    def from_lxml_element(cls, element, clone=False) -> XmlNode:
        '''
            Create an `XmlNode` from an `lxml` element.

            Arguments:
                element: `lxml` element
        '''
        if clone:
            return XmlNode(element)
        else:
            return XmlNode(element).clone()

    @classmethod
    def node_locator(cls, *, tags: 'strOrSequence'=None, **attrs):
        '''
            Create a locator for finding an XML Node in an **XmlNode**.

            Keyword Arguments:
                tags: (Optional) Descendant tags for the node. Can be a string of single or multiple tags or a list/tuple of tags.
                **attrs: Arbitrary number of key value pairs representing attribute name and value.

            Raises:
                Exception: If neither tag nor an attribute is provided.

            Note:
                You can use tag and attributes in combination.

                Supports nested node finding.
        '''
        return NodeLocator(tags=tags, **attrs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Arjuna

Arjuna is a Python based test automation framework developed by Rahul Verma (www.rahulverma.net). Rahul has implemented smaller variants of features in Arjuna across frameworks and organizations, or given advise around it. However Arjuna being a generic library has the most complete implementation of his ideas, away from project specific contexts and constraints.

Arjuna uses **[pytest](https://docs.pytest.org/en/latest/)** as its recommended test engine. Arjuna also provides its markup for some common use cases on top of pytest. If you like, you can make use of Arjuna with any other test engine or custom frameworks as well.

To import all public names for test automation authors with
from arjuna import *

To run arjuna as a module

    $python -m arjuna <options>

To see which command line options are available:

    $python -m arjuna -h

All classes, functions and names supposed to be used in test automation code are available in Tester Programming Interface: **arjuna.tpi** package.
"""

import pkg_resources
import time
import os
import sys

def __join_paths(*paths):
    return os.path.abspath(os.path.join(*paths))

__root_dir = __join_paths(os.path.dirname(os.path.realpath(__file__)), "..")
__importables_dir = __join_paths(__root_dir, "third_party")

sys.path.insert(0, __importables_dir)
sys.path.insert(0, __root_dir)

class _ArjunFacade:

    def __init__(self):
        self.__version = pkg_resources.require("arjuna")[0].version

    def load(self, raw_args):
        from arjuna.tpi.parser.text import _TextResource
        from arjuna.interface.cli import ArjunaCLI
        reader = _TextResource("header.txt")
        print(reader.read().format(version=self.__version))
        reader.close()

        print("Parsing CLI Options...")
        cli = ArjunaCLI(raw_args)
        # Initialize the Arjuna Core as per CLI options
        print("Intializing Arjuna...")
        cli.init()

        print("Loading Arjuna core to allow third party test engine to take over test execution...")
        cli.load()

    def launch(self, raw_args):
        from arjuna.tpi.parser.text import _TextResource
        from arjuna.interface.cli import ArjunaCLI
        reader = _TextResource("header.txt")
        print(reader.read().format(version=self.__version))
        reader.close()

        print("Parsing CLI Options...")
        cli = ArjunaCLI(raw_args)
        # Initialize the Arjuna Core as per CLI options
        print("Intializing Arjuna...")
        cli.init()

        print("Executing Arjuna Command...")
        cli.execute()

from arjuna.tpi.engine import Arjuna

from arjuna.tpi.error import *
from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.tpi.tracker import track
from arjuna.tpi.helper.arjtype import nvpair, nvpairs, withx, Screen, attr, node, bnode, fnode, Point, Offset, NetworkPacketInfo, oneof, axes
from arjuna.tpi.helper.extract import pos
from arjuna.tpi.helper.audit import HardCoded
from arjuna.tpi.helper.arjtype import ProcessedKeyDict
from arjuna.tpi.helper.datetime import Time, DateTime, DateTimeDelta, DateTimeStepper

from arjuna.tpi.parser.json import Json, JsonSchema
from arjuna.tpi.parser.xml import Xml
from arjuna.tpi.parser.html import Html
from arjuna.tpi.parser.text import Text

from arjuna.tpi.engine.test import *
from arjuna.tpi.engine.relation import *
from arjuna.tpi.engine.resource import *
from arjuna.tpi.engine.data_markup import *

from arjuna.tpi.error import *

from arjuna.tpi.data.record import *
from arjuna.tpi.data.generator import Random, Locales, generator, composite, composer
from arjuna.tpi.data.entity import data_entity

from arjuna.tpi.guiauto.meta.locator import GuiWidgetDefinition
gui_widget_def = GuiWidgetDefinition
widget = GuiWidgetDefinition
from arjuna.tpi.guiauto.model.app import GuiApp
from arjuna.tpi.guiauto.model.page import GuiPage
from arjuna.tpi.guiauto.model.section import GuiSection
from arjuna.tpi.guiauto.model.dialog import GuiDialog
from arjuna.tpi.guiauto.helper import Keys

from arjuna.tpi.httpauto.http import Http
from arjuna.tpi.httpauto.oauth import OAuthImplicitGrantSession

from arjuna.tpi.dbauto.db import DB

from arjuna.tpi.hook.config import Configurator

from arjuna.tpi.log import *
from arjuna.tpi.magic import C, L, R



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import inspect
import os

class _Stack:

    @classmethod
    def get_invoker(cls):
        frame = inspect.stack()[2]
        mod = inspect.getmodule(frame[0])
        mod_file = frame.filename
        mod_name = os.path.basename(mod_file).split(".")[0]
            
        mod_script = ""
        if mod is not None and mod.__name__ == "__main__":
            mod_script = "Script:<{}> at ".format(mod.__file__)
        else:
            mod_script = "Module:<{}> File:<{}>".format(mod_name, mod_file)
        func = frame[3]
        if func == "<module>":
            func = ""
        else:
            func = "Function/Method: <{}> in ".format(func)
        line = frame[2]
        return "{}{}Line: {}".format(func, mod_script, line)

class HardCoded:
    '''
        It's sole purpose is to increase responsbility of test author when there is a need for **time.sleep**
    '''

    @classmethod
    def __log(cls, invoker, why, seconds):
        from arjuna import log_warning
        log_warning("Hardcoded sleep executed for {} seconds by {}. Reason by author: {}".format(seconds, invoker, why))

    @classmethod
    def sleep(cls, why: str, seconds: float) -> None:
        '''
            Fixed/Static sleep.

            Logs a warning message mentioning the caller, reason and number of seconds.
            Arguments:
                why: Reason for using static wait instead of dynamic waits that Arjuna provides.
                seconds: Number of seconds for sleeping.
        '''
        time.sleep(seconds)
        cls.__log(_Stack.get_invoker(), why, seconds)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import argparse
from functools import partial

from .validation import *
from arjuna.core.constant import ReportFormat


class Parser:

    def __init__(self):
        self.parser = None

    def get_parser(self):
        return self.parser

class ProjectParser(Parser):

    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('--project', metavar="test_project_root_dir", dest="project.root.dir", type=project_dir, help = 'Valid absolute path of an existing Arjuna test project. Needed with python -m arjuna call only. With python arjuna_launcher.py you can skip this argument as the script determines its container Arjuna test project automatically.')

    def process(self, arg_dict):
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class NewProjectParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('--project', metavar="test_project_root_dir", dest="project.root.dir", type=new_project_dir, help = 'Absolute non-existing Project root directory. Last part of directory represents project name. Name of project should be an Alnum 3-30 length with only lower case letters.')

    def process(self, arg_dict):
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


import sys
import os
import pytest
import threading
import platform
import pkg_resources

from arjuna.engine.pytest import PytestHooks

from functools import partial

from arjuna.interface.cli.validation import *
from arjuna.interface.enums import TargetEnum
from arjuna.core.error import *
from arjuna.tpi.error import *
from arjuna.tpi.parser.yaml import Yaml


_ARJUNA_CLI_ARGS = {
    "project": ("--project", {
        "dest":"project", 
        "metavar":"project_root_dir", 
        "type": project_dir,
        "help":'Valid absolute path of an existing Arjuna test project. If not provided pytest rootdir value is considered. If that is not provided, current working directory is considered.'
    }),

    "run.id": ("--rid", {
        "dest":"run.id", 
        "metavar":"run_id", 
        # "type":partial(lname_check, "Run ID"), 
        "help": 'Alnum 3-30 length. Only lower case letters.', 
        # "default":"mrun"
    }),

    "report.formats": ('--otype', {
        "dest":"report.formats", 
        "type":report_format, 
        "action":"append", 
        "choices":[i for i in ReportFormat.__members__], 
        "help":'Output/Report format. Can pass any number of these switches.',
        # "default": ['XML', 'HTML']
     }), # "choices":['XML', 'HTML'], 

    "link.projects": ('--link', {
        "dest":"link.projects", 
        "action":"append", 
        "help":'Linked Arjuna Test Project.',
    }),

    "static.rid": ('--update', {
        "dest":"static.rid", 
        "action":'store_true', 
        "help": 'Will result in overwriting of report files. Useful during script development.'
    }),

    "dry.run": ('--dry-run', {
        "dest":"dry.run", 
        "choices":[i for i in DryRunType.__members__], 
        "type":ustr, 
        "help":'Does a dry run. Tests are not executed. Behavior depends on the type passed as argument. SHOW_TESTS - enumerate tests. SHOW_PLAN - enumerates tests and fixtures. RUN_FIXTURES - Executes setup/teardown fixtures and emuerates tests.'
    }),

    "ref.conf": ('--rconf', {
        "dest":"ref.conf", 
        "metavar":"config_name", 
        # "type":str, 
        "help":"Run/Reference Configuration object name for this run. Default is 'ref'",
    }),

    "ao": ('--ao', {
        "dest":"ao",
        "nargs":2,
        "action":'append',
        "metavar":('option', 'value'),
        "help":'Arjuna Option. Can pass any number of these switches.'
    }),

    "uo": ("--uo", {
        "dest":"uo",
        "nargs":2,
        "action":'append',
        "metavar":('option', 'value'),
        "help":'User Option. Can pass any number of these switches.'
    }),

    'log.console.level': ('--display-level', {
        "dest":'log.console.level', 
        "type":ustr, 
        "choices":[i for i in LoggingLevel.__members__],
        "help":"Minimum message level for display. (choose from 'DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL')", 
        # "default":LoggingLevel.INFO.name
    }),

    'log.file.level': ('--logger-level', {
        "dest":'log.file.level', 
        "type":ustr, 
        "choices":[i for i in LoggingLevel.__members__],
        "help":"Minimum message level for log file. (choose from 'DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL')", 
        # "default":LoggingLevel.DEBUG.name
    }),

    "group": ("--group", {
        "dest": "group", 
        "metavar": "group_name", 
        "type": partial(lname_check, "Group Name"), 
        "help": 'Name of a defined group in test groups configuration file in <Project Root>/config/groups.yaml file.',
        "default": "mgroup"
    }),

    'ipack': ('--ipack', {
        "dest":"ipack", 
        "action":"append", 
        "metavar":"package_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for including test packages. Can pass any number of these switches.'
    }),

    'epack': ('--epack', {
        "dest":"epack", 
        "action":"append", 
        "metavar":"package_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for excluding test packages. Can pass any number of these switches.'
    }),

    'imod': ('--imod', {
        "dest":"imod", 
        "action":"append", 
        "metavar":"module_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for including test modules. Can pass any number of these switches.'
    }),

    'emod': ('--emod', {
        "dest":"emod", 
        "action":"append", 
        "metavar":"module_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for excluding test modules. Can pass any number of these switches.'
    }),

    'itest': ('--itest', {
        "dest":"itest", 
        "action":"append", 
        "metavar":"test_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for including test functions. Can pass any number of these switches.'
    }),

    'etest': ('--etest', {
        "dest":"etest", 
        "action":"append", 
        "metavar":"test_name_or_regex", 
        "default":None, 
        "help":'Name/regex pattern for excluding test functions. Can pass any number of these switches.'
    }),

    'irule': ('--irule', {
        "dest":"irule", 
        "action":"append", 
        "metavar":"rule", 
        "default":None, 
        "help":'Include test functions that match the rule. Can pass any number of these switches. Test Function is included if any of the inclusion rules matches.'
    }),

    'erule': ('--erule', {
        "dest":"erule", 
        "action":"append", 
        "metavar":"rule", 
        "default":None, 
        "help":'Exclude test functions that match the rule. Can pass any number of these switches. Test Function is excluded if any of the exclusion rules matches. Evaluated before any inclusion rules.'
    }),
}

def convert_to_absolute(p):
    if p.startswith("./test") or p.startswith(".\\test"):
        return os.path.join(pytest_root_dir + p[1:])
    else:
        return p

def pytest_addoption(parser):
    from arjuna.tpi.parser.text import _TextResource
    reader = _TextResource("header.txt")
    print(reader.read().format(version=pkg_resources.require("arjuna")[0].version))
    reader.close()

    group = parser.getgroup("arjuna", "Arjuna Test Automation Framework")
    for k,v in _ARJUNA_CLI_ARGS.items():
        group.addoption(v[0], **v[1])

_commands = {
    "PROJECT" : "run-project",
    "GROUP" : "run-group",
    "SELECTED" : "run-selected",
}

RAW_ARGS = ["pytest"]
CONVERTED_ARGS = ["pytest", "-p", "arjuna"]

def _determine_project_root_dir(args):
    def process_rootdir_or_project(index, arg, oargs):
        argparts = [p.strip() for p in arg.split("=")]
        if len(argparts) == 2:
            project_path = argparts[1]
            oargs[index] = "__REMOVE__"
        else:
            project_path = args[index + 1]
            if project_path.startswith("-"):
                raise Exception("For rootdir switch you must provide a path.")

            oargs[index] = "__REMOVE__"
            oargs[index + 1] = "__REMOVE__"
        return project_path

    project_path = None
    root_dir = None

    from copy import deepcopy
    targs = deepcopy(args)

    for index, arg in enumerate(targs):
        if arg.lower().startswith("--rootdir"):
            root_dir = process_rootdir_or_project(index, arg, args)
        elif arg.lower().startswith("--project"):
            project_path = process_rootdir_or_project(index, arg, args)

    if project_path is None:
        pytest_root_dir = os.getcwd()
    elif project_path is not None:
        if project_path == ".":
            pytest_root_dir = os.getcwd()
        else:
            pytest_root_dir = project_path
    elif root_dir is not None:
        pytest_root_dir = root_dir

    if not os.path.isabs(pytest_root_dir):
        pytest_root_dir = os.path.join(os.getcwd(), pytest_root_dir)

    CONVERTED_ARGS.extend(['--project', pytest_root_dir])

    args[:] = [e for e in args if e != "__REMOVE__"]

    args[:] = [convert_to_absolute(e) for e in args]
    sys.path.append(pytest_root_dir + "/..")
    args.extend(['--rootdir=' + pytest_root_dir])
    args.extend(['-c', pytest_root_dir + "/pytest.ini"])

    conf_file = os.path.join(pytest_root_dir, "test/conftest.py")
    if os.path.exists(conf_file):
        os.remove(conf_file)


    return pytest_root_dir

def _add_pytest_addl_args(args):
    res_path = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../res"))
    # pytest_ini_path = res_path + "/pytest.ini"
    css_path = res_path + "/arjuna.css"
    more_args = ["--disable-warnings", "-rxX", "--css", css_path, "--log-level", "FATAL"] #"-c", pytest_ini_path, 
    args.extend(more_args)
    CONVERTED_ARGS.extend(more_args)

    platform_args = None
    if platform.system().casefold() == "Windows".casefold() :
        platform_args = ["--capture", "no"]
    else:
        platform_args = ["--show-capture", "all"] #, "--no-print-logs"]
    args.extend(platform_args)
    CONVERTED_ARGS.extend(platform_args)    

def _handle_dry_run_option(args):
    dry_run_raw = None
    for index, arg in enumerate(args):
        if arg.lower().startswith("--dry-run"):
            argparts = [p.strip() for p in arg.split("=")]
            if len(argparts) == 2:
                dry_run_raw = argparts[1]
            else:
                dry_run_raw = args[index + 1]
                if dry_run_raw.startswith("-"):
                    raise Exception("For --dry-run switch you must provide a dry run type.")
            break

    if dry_run_raw is not None:
        from arjuna.core.constant import DryRunType
        try:
            dry_run = DryRunType[dry_run_raw.upper()]
        except Exception as e:
            raise Exception("Invalid dry run type. Should be one of show_tests/show_plan/create_res")

        print("!!!!!! This is a DRY RUN !!!!!!!")
        if dry_run == DryRunType.SHOW_TESTS:
            print("Dry Run Type: SHOW TESTS")
            print("You can see the test functions which will be executed as per settings of your command.")
            darg = "--collect-only"
        elif dry_run == DryRunType.SHOW_PLAN:
            print("Dry Run Type: SHOW PLAN")
            print("You can see the test functions as well as the fixtures which will be executed as per settings of your command.")
            darg = "--setup-plan"
        elif dry_run == DryRunType.CREATE_RES:
            print("Dry Run Type: CREATE RESOURCES")
            print("All resources will be created as per your current command. You can see the test functions which will be executed as per settings of your command.")
            darg = "--setup-only"

        args.extend([darg])
        CONVERTED_ARGS.extend([darg]) 


# def pytest_cmdline_parse(pluginmanager, args):
@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(early_config, parser, args):
    # print(early_config.known_args_namespace)
    # sys.exit()
    if '--help' in args or '-h' in args:
        return
    RAW_ARGS.extend([arg.find(" ") == -1 and arg or f'"{arg}"' for arg in args])
    pytest_root_dir = _determine_project_root_dir(args)

    res_path = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../res"))
    # pytest_ini_path = res_path + "/pytest.ini"
    css_path = res_path + "/arjuna.css"
    import pathlib
    # early_config.inifile = pathlib.Path(pytest_ini_path)

    _add_pytest_addl_args(args)
    _handle_dry_run_option(args)
    global pytestargs
    pytestargs = args

_ONLY_CMD = {"project", "run.id", "static.rid", "link.projects", "dry.run", "ref.conf", "group"}

def _load_argdict_rule_dict(config):
    rule_dict = dict()
    arg_dict = {}
    for option in _ARJUNA_CLI_ARGS:
        value = config.getoption(option)
        if option.lower() in _ONLY_CMD:
            continue
        if option.lower() in {'ipack', 'epack', 'imod', 'emod', 'itest', 'etest', 'irule', 'erule'}:
            rule_dict[option.lower()] = value
            if value:
                for v in value:
                    if option.lower() in {"irule", "erule"}:
                        CONVERTED_ARGS.extend(["--" + option, f'"{v}"'])    
                    else:
                        CONVERTED_ARGS.extend(["--" + option, v])
            continue
        if value:
            arg_dict[option] = value
    return arg_dict, rule_dict

def _init_arjuna(config, arg_dict):
    from arjuna import Arjuna
    from arjuna.configure.cli import CliArgsConfig
    cliconfig = CliArgsConfig(arg_dict)

    linked_projects = config.getoption("link.projects")
    if linked_projects is not None:
        out = []
        for lp in linked_projects:
            if not os.path.isabs(lp):
                out.append(os.path.join(os.getcwd(), lp))
            else:
                out.append(lp)
        linked_projects = out
    Arjuna.init(
        config.option.rootdir, 
        run_id=config.getoption("run.id"), 
        static_rid=config.getoption("static.rid"), 
        linked_projects=linked_projects, 
        arjuna_options=cliconfig.arjuna_options, 
        user_options=cliconfig.user_options
    )

    from arjuna import C
    os.chdir(C("project.root.dir"))

class TestGroup:

    def __init__(self, *, name="mgroup", rconf_name="ref"):
        from arjuna import Arjuna
        from arjuna.tpi.constant import ArjunaOption
        Arjuna.register_group_params(name=name, config=Arjuna.get_config(rconf_name), thread_name=threading.currentThread().name)    

    def __create_rule_strs(self, include_exclude_dict):
        pickers_rulestr = {
            'ipack': "package *= {}",
            'epack': "package *= {}",
            'imod': "module *= {}",
            'emod': "module *= {}",
            'itest': "name *= {}",
            'etest': "name *= {}",
        }

        rules = {'ir': [], 'er': []}

        for picker in pickers_rulestr:
            names = include_exclude_dict.pop(picker)
            if names:
                for name in names:
                    if picker.startswith('i'):
                        rules['ir'].append(pickers_rulestr[picker].format(name))
                    else:
                        rules['er'].append(pickers_rulestr[picker].format(name))

        for k,v in include_exclude_dict.items():
            if not v: continue
            if k == "irule":
                rules['ir'].extend(v)
            elif k == "erule":
                rules['er'].extend(v)

        return rules

    def _load_tests(self, rule_dict):
        i_e_rules = self.__create_rule_strs(rule_dict)
        rules = {'ir': [], 'er': []}
        rules['ir'].extend(i_e_rules['ir'])
        rules['er'].extend(i_e_rules['er'])

        from arjuna.engine.selection.selector import Selector
        selector = Selector()
        if rules:
            for rule in rules['ir']:
                selector.include(rule)
            for rule in rules['er']:
                selector.exclude(rule)

        from arjuna import Arjuna
        Arjuna.register_test_selector_for_group(selector)

    @classmethod
    def from_pickers(cls, *, rconf_name=None, rules={}):
        if rconf_name is None: rconf_name = "ref"
        group = TestGroup(rconf_name=rconf_name)
        group._load_tests(rules)
        return group

    @classmethod
    def __get_group_yaml(cls, name):
        from arjuna import C
        gfile = C(ArjunaOption.CONF_GROUPS_LOCAL_FILE)
        if not os.path.isfile(gfile):
            gfile = C(ArjunaOption.CONF_GROUPS_FILE)                
        try:
            gyaml = Yaml.from_file(gfile, allow_any=True)
        except FileNotFoundError:
            raise TestGroupsFileNotFoundError(file_path=gfile)
           
        try:
            return gyaml.get_section(name)
        except YamlUndefinedSectionError as e:
            raise UndefinedTestGroupError(name=name, file_path=gfile)

    @classmethod
    def from_def(cls, name, *, rconf_name=None, rules={}):
        group_yaml = cls.__get_group_yaml(name)
        from arjuna import Arjuna
        for gmd_name in group_yaml.section_names:
            rules = {
                'ipack': None,
                'epack': None,
                'imod': None,
                'emod': None,
                'itest': None,
                'etest': None,
                'irule' : None,
                'erule': None
            }
            gmd_name = gmd_name.lower()
            if gmd_name == "conf":
                rconf_name = group_yaml["conf"]
            elif gmd_name in rules:
                rules[gmd_name] = group_yaml[gmd_name]

        if rconf_name is None:
            rconf_name = "ref"
        group = TestGroup(name=name, rconf_name=rconf_name)
        group._load_tests(rules)
        return group

def _configure_pytest_reports(config):
    from arjuna import Arjuna
    xml_path = os.path.join(Arjuna.get_config().value(ArjunaOption.REPORT_XML_DIR), "report.xml")
    html_path = os.path.join(Arjuna.get_config().value(ArjunaOption.REPORT_HTML_DIR), "report.html")
    report_formats = Arjuna.get_config().value(ArjunaOption.REPORT_FORMATS)

    if ReportFormat.XML in report_formats:
        if not config.option.xmlpath:
            config.option.xmlpath = xml_path
            CONVERTED_ARGS.extend(["--junit-xml", xml_path])

    if ReportFormat.HTML in report_formats:
        config.option.htmlpath = html_path
        config.option.self_contained_html = True
        CONVERTED_ARGS.extend(["--html", html_path])
        CONVERTED_ARGS.extend(["--self-contained-html"])

@pytest.hookimpl(tryfirst=True)
def pytest_configure(config):
    if config.getoption("help"):
        return

    arg_dict, rule_dict = _load_argdict_rule_dict(config)
    _init_arjuna(config, arg_dict)
    gname = config.getoption("group").lower()
    if gname == "mgroup":
        group = TestGroup.from_pickers(rconf_name=config.getoption("ref.conf"), rules=rule_dict)
    else:
        if rule_dict:
            raise Exception("When you provide a group name, you can provide test selectors/rules in Command line. You have provided group name as >>{}<< and also provided selectors/rules: {}".format(gname, {k:v for k,v in rule_dict.items() if v}))
        group = TestGroup.from_def(gname, rconf_name=config.getoption("ref.conf"))

    _configure_pytest_reports(config)

    from arjuna import Arjuna
    Arjuna._set_command(" ".join(RAW_ARGS))
    Arjuna.register_pytest_command_for_group(" ".join(CONVERTED_ARGS))

    PytestHooks.add_env_data(config)

def pytest_collection_modifyitems(items, config):
    # from arjuna import Arjuna
    # session = Arjuna.get_test_session()
    # session.load_tests(dry_run=config.getoption("dry.run"), ref_conf_name=config.getoption("ref.conf"))
    PytestHooks.select_tests(items, config)

def pytest_generate_tests(metafunc):
    PytestHooks.configure_group_for_test(metafunc)

@pytest.mark.hookwrapper
def pytest_runtest_makereport(item, call):
    result = yield
    PytestHooks.prepare_result(result)
    PytestHooks.enhance_reports(item, result)

def pytest_html_report_title(report):
    PytestHooks.set_report_title(report)

def pytest_html_results_summary(prefix, summary, postfix):
    PytestHooks.inject_arjuna_js(prefix)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.error import *
from arjuna.core.error import *

class NamedString:
    def __init__(self, internal, external):
        self.code = internal
        self.name = external

    def get_code(self):
        return self.code

    def get_name(self):
        return self.name


class NamedStringsContainer:
    def __init__(self, name):
        self.container_name = name
        self.names = []

    def add(self, name):
        self.names.append(name)

    def get_name(self):
        return self.container_name

    def get_named_strings(self):
        return self.names


class Name(NamedString):
    pass


class Message(NamedString):
    pass


class NamesContainer(NamedStringsContainer):
    pass


class MessagesContainer(NamedStringsContainer):
    pass


def populate_from_codelist(map, codes):
    for index, code in enumerate(codes):
        map[code[0].upper().trim()] = code[2]


def populate_from_codemap(map, codes):
    for k, v in codes:
        map[k.upper()] = v


def populate_from_namedstringlist(map, named_strings):
    for ns in named_strings:
        map[ns.get_code().upper()] = ns.get_name()


class StringsManager:
    def __init__(self):
        self.msg_map = {}
        self.name_map = {}
        self.prop_map = {}
        self.flattened_names = {}

    def populate_names(self, names_containers_list):
        for nc in names_containers_list:
            if nc.get_name() not in self.name_map:
                self.name_map[nc.get_name()] = {}
            populate_from_namedstringlist(self.name_map[nc.get_name()], nc.get_named_strings())

    def populate_messages(self, messages_containers_list):
        for mc in messages_containers_list:
            if mc.get_name() not in self.msg_map:
                self.msg_map[mc.get_name()] = {}
            populate_from_namedstringlist(self.msg_map[mc.get_name()], mc.get_named_strings())

    def populate_flattened_names(self):
        for section in self.name_map:
            for key in self.name_map[section]:
                self.flattened_names[section + "::" + key] = self.name_map[section][key]

    def get_all_names(self):
        return self.name_map

    def get_all_messages(self):
        return self.msg_map;

    def get_flattneded_names(self):
        return self.flattened_names;

    def __section_exists(self, section):
        return section in self.msg_map

    def __code_exists(self, section, code):
        if not self.__section_exists(section):
            return False;
        return code in self.msg_map[section]

    def __throw_not_initialized_exception(self, context, method):
        raise Problem("adv", context, method, "LOCALIZER_NOT_INITIALIZED",
                      "Strings Manager not initialized.")

    def __get_text_for_code(self, section, msg_code):
        section_code = section.to_upper_case().trim();
        code = msg_code.to_upper_case().trim();
        if not self.__code_exists(section_code, code):
            return code;
        return self.msg_map[section_code][code]

    def get_info_message_text(self, msg_code):
        return self.__get_text_for_code("INFO_MESSAGES", msg_code)

    def get_problem_text(self, msg_code):
        return self.__get_text_for_code("PROBLEM_MESSAGES", msg_code)

    def get_warning_text(self, msg_code):
        return self.__get_text_for_code("WARNING_MESSAGES", msg_code);

    def get_configured_name(self, section_name, internal_name):
        return self.name_map[section_name.to_upper_case().trim()][internal_name.to_upper_case().trim()]

    class problem_codes:
        pass

    class info_codes:
        pass

    class error_codes:
        pass

    def add_property_name(self, prop_code, prop_name):
        self.prop_map[prop_code.upper()] = prop_name

    def get_property_name(self, prop_code):
        return self.prop_map[prop_code.upper()]



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

class Asserter:
    '''
        Arjuna's asserter class.

        It can be used directly. It is already included as an attribute for request fixture in test functions and all Guis and Gui Templates.
    '''

    def __init__(self):
        self.__asserter = unittest.TestCase('__init__')

    @classmethod
    def _format_msg(cls, msg):
        return msg and " {}.".format(msg) or ""

    def assert_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 == obj2

            Wrapper on unittest's assertEqual

            Args:
                obj1: Object of any type which supports == operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertEqual(obj1, obj2, msg)

    def assert_approx_equal(self, obj1, obj2, msg, places=None, delta=None):
        '''
            Assert round(obj1-obj2, places) == 0 if places is supplied

            Assert round(obj1-obj2, places) == delta if delta is supplied

            Wrapper on unittest's assertAlmostEqual

            Args:
                obj1: Object of any type which supports == operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
                places: Number of decimal places used for rounding
                delta: The difference between objects after rounding (diff <= delta for passing assertion)

            Note:
                You can specify either places or delta not both.

            Note:
                This method rounds the values to the given number of decimal places (i.e. like the Python's round() function) and not significant digits.
        '''
        if places is not None and delta is not None:
            raise TypeError("Specify delta or places not both.")
        elif places is None and delta is None:
            self.__asserter.assertAlmostEqual(obj1, obj2, msg=msg)
        elif places is None:
            self.__asserter.assertAlmostEqual(obj1, obj2, msg=msg, delta=delta)
        else:
            self.__asserter.assertAlmostEqual(obj1, obj2, msg=msg, places=places)

    def assert_approx_not_equal(self, obj1, obj2, msg, places=7, delta=None):
        '''
            Assert round(obj1-obj2, places) != 0 if places is supplied

            Assert round(obj1-obj2, places) > delta if delta is supplied

            Wrapper on unittest's assertNotAlmostEqual

            Args:
                obj1: Object of any type which supports == operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
                places: Number of decimal places used for rounding
                delta: The allowed difference between objects after rounding (diff >= delta for passing assertion)

            Note:
                You can specify either places or delta not both.

            Note:
                This method rounds the values to the given number of decimal places (i.e. like the Python's round() function) and not significant digits.
        '''

        if places is not None and delta is not None:
            raise TypeError("Specify delta or places not both.")
        elif places is None and delta is None:
            self.__asserter.assertNotAlmostEqual(obj1, obj2, msg=msg)
        elif places is None:
            self.__asserter.assertNotAlmostEqual(obj1, obj2, msg=msg, delta=delta)
        else:
            self.__asserter.assertNotAlmostEqual(obj1, obj2, msg=msg, places=places)

    def assert_exceptions(self, exceptions, callable=None, *cargs, regex=None, **ckwargs):
        '''
            Asserts that one of the provided exceptions is raised when the provided callable (or ) is called with provided arguments.

            Args:
                exceptions: One or more Exception classes provided as tuple. A single Exception can be directly provided.
                callable: The callable to be called.
                cargs: Positional arguments to be passed to callable

            Keyword Arguments:
                regex: Regular expression to match the string representation of the Exception. Can be a string or a regular expression object that can be used with Python's **re.search()** call.
                ckwargs: Keyword arguments to be passed to callable

            Note:
                You can use this method to create a context manager by not providing the callable.

                This means you can write the code to be asserted for directly inline:
                
                    .. code-block:: python

                        with asserter.assert_exceptions(SomeException):
                            do_something()

                If you want to futher put assertions on the the actual exception raised, you can use context manager's **exception** attribute:

                    .. code-block:: python

                        with asserter.assert_exceptions(SomeException) as cm:
                            do_something()

                        asserter.assert_equal(cm.exception.error_code, 3)
        '''
        if callable is None:
            if cargs or ckwargs:
                raise Exception("Positional and/or Keywrod arguments are allowed only when callable is provided.")
            if regex is not None:
                return self.__asserter.assertRaisesRegex(exceptions, regex)
            else:
                return self.__asserter.assertRaises(exceptions)
        else:
            if regex is not None:
                self.__asserter.assertRaisesRegex(exceptions, regex, callable, *cargs, **ckwargs)
            else:
                return self.__asserter.assertRaises(exceptions, callable, *cargs, **ckwargs)
                
            


    def assert_lesser(self, obj1, obj2, msg):
        '''
            Assert obj1 < obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports < operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLess(obj1, obj2, msg)

    def assert_greater(self, obj1, obj2, msg):
        '''
            Assert obj1 > obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports > operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreater(obj1, obj2, msg)

    def assert_min(self, obj, min_value, msg):
        '''
            Asserts a minimum value for an object i.e. obj >= min_value

            Wrapper on unittest's assertGreaterEqual

            Args:
                obj: Object of any type which supports >= operator for min_value
                min_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreaterEqual(obj, min_value, msg)

    def assert_max(self, obj, max_value, msg):
        '''
            Asserts a maximum value for an object i.e. obj <= max_value

            Wrapper on unittest's assertLessEqual

            Args:
                obj: Object of any type which supports <= operator for min_value
                max_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLessEqual(obj, max_value, msg)

    def assert_not_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 != obj2

            Wrapper on unittest's assertNotEqual

            Args:
                obj1: Object of any type which supports != operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertNotEqual(obj1, obj2, msg)

    def assert_true(self, obj, msg):
        '''
            Assert obj is True.

            Wrapper on unittest's assertTrue

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertTrue(obj, msg)

    def assert_false(self, obj, msg):
        '''
            Assert obj is False.

            Wrapper on unittest's assertFalse

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertFalse(obj, msg)

    def fail(self, msg):
        '''
            Raises AssertionError with the provided message.

            Args:
                msg: A context string explaining the failure.
        '''
        self.__asserter.fail(msg)


class AsserterMixIn:
    '''
        Base class to add **asserter** property to any class which inherits from it.
    '''

    def __init__(self):
        # Trick to use assertions outside of a unittest test
        self.__asserter = Asserter()

    @property
    def asserter(self) -> Asserter:
        '''
            Arjuna's Asserter object for executing assertions.
        '''
        return self.__asserter

class IterableAsserterMixin:

    '''
        Base class for adding assertions to iterables.

        Note:
            Use it in combination with AsserterMixIn.
            The child class should have `_container` as its attribute or property.
    '''

    def __init__(self):
        self.__klass_name = self.__class__.__name__

    def __len__(self):
        return len(self._container)

    def is_empty(self):
        '''
            Returns True if this iterable has no objects.
        '''
        return len(self) == 0

    def assert_empty(self, *, msg):
        '''
            Assert that there should not be any object in this iterable.

            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if not self.is_empty():
            raise AssertionError("{} is not empty. Length: {}. {}".format(self.__klass_name, len(self), msg))

    def assert_not_empty(self, *, msg):
        '''
            Assert that there should be atleast one object in this iterable.

            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if self.is_empty():
            raise AssertionError(f"{self.__klass_name} is empty. {msg}")

    def assert_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable should match provided size.

            Arguments:
                size: Expected number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length != size:
            raise AssertionError(f"{self.__klass_name} is not of expected size. Expected: {size}. Actual: {length}. {msg}")

    def assert_min_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable >= provided size.

            Arguments:
                size: Expected minimum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length < size:
            raise AssertionError(f"{self.__klass_name} is not of minimum expected size. Expected minimum size is {size}. Actual: {length}. {msg}")

    def assert_max_size(self, size, *, msg):
        '''
            Assert that the number of objects in this iterable <= provided size.

            Arguments:
                size: Expected maximum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        if size <0 :
            raise TypeError("Negative number aren't allowed")
        length = len(self)
        if length > size:
            raise AssertionError(f"{self.__klass_name} is not of maximum expected size. Expected maximum size is {size}. Actual: {length}. {msg}")

    def assert_size_range(self, min_size, max_size, *, msg):
        '''
            Assert that the number of objects in this iterable should be in the provided size range.

            Arguments:
                min_size: Expected minimum number of objects.
                max_size: Expected maximum number of objects.
            
            Keyword Arguments:
                msg: Purpose of this assertion.
        '''
        length = len(self)
        if length < min_size or length > max_size:
            raise AssertionError(f"{self.__klass_name} is not in expected size range. Expected min: {min_size} < actual: {length} < max: {max_size}. {msg}")







# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.engine.asserter import AsserterMixIn
from arjuna.tpi.helper.arjtype import NotFound
from arjuna.tpi.parser.text import Text

class Validator(AsserterMixIn):

    def __init__(self, response):
        super().__init__()
        self.__response = response

    @property
    def response(self):
        return self.__response

    def validate(self, name, target):
        stored_value = self.__response.store[name]

        def validate_found(expected_to_be_found):
            exists = True
            if isinstance(stored_value, NotFound):
                exists = False
            if expected_to_be_found:
                self.asserter.assert_true(exists, f"No value was extracted for name >>{name}<<.")
            else:
                self.asserter.assert_false(exists, ">>{}<< was extracted for name >>{}<<. It should not exist.".format(stored_value, name))

        if "exists" not in target:
            validate_found(True)
        else:
            validate_found(target['exists'])
            # No further validations as it is not found
            if not target['exists']:
                return

        for k,v in target.items():
            k = k.lower()
            if k == "exists":
                continue
            elif k == "contains":
                proc_stored_value = stored_value
                if type(stored_value) is str:
                    proc_stored_value = Text(stored_value)
                for item in v:
                    getattr(proc_stored_value, "assert_contains")(item, msg=f"Extracted value for >>{name}<< does not contain >>{item}<<.")
            elif k == "min":
                self.asserter.assert_min(stored_value, v, msg=f"Extracted value for >>{name}<< did not match min value criterion.")





    


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
import json

from arjuna.tpi.constant import ArjunaOption
from arjuna.core.har import Har
from arjuna.tpi.protocol.network_recorder import NetworkRecorder
from arjuna.tpi.tracker import track

@track("trace")
class BrowserMobNetworkRecorder(NetworkRecorder):

    def __init__(self, automator):
        super().__init__()
        self.__automator = automator
        self.__start()

    @property
    def _proxy(self):
        return self.__automator.dispatcher.proxy

    def __new_har(self):
        # from arjuna.tpi.data.generator import Random
        # temp_dir = self.__automator.config.value(ArjunaOption.TEMP_DIR)
        # har_file_path = os.path.join(temp_dir, "{}.har".format(Random.ustr()))
        self._proxy.new_har(options={"captureHeaders":True,"captureCookies":True,"captureContent":True})

    def _is_active(self):
        return self._proxy is not None

    def __start(self):
        if self._proxy is None: return
        self.__new_har()
        self._proxy.new_page(title="Ignore")

    def _record(self, title):
        self.__new_har()
        self._proxy.new_page(title=title)

    @property
    def har(self):
        if not self._is_active(): return None
        return self._proxy.har

    def _register(self):
        from arjuna import Arjuna, ArjunaOption
        har = Har.from_har(self.har)
        filter_resources = self.__automator.config.value(ArjunaOption.REPORT_NETWORK_FILTER)
        for packet in har.get_pagewise_network_packet_info(filter_resources=filter_resources):
            Arjuna.get_report_metadata().add_network_packet_info(packet)
        self.__new_har() # This is done to clear har.
        self._proxy.new_page(title="Ignore")




# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum, auto
from collections import namedtuple

from .meta import Meta
from .enums import WithType
from ._with import With, ImplWith, Locator, GuiGenericLocator
from .translator import LocatorTranslator
from arjuna.core.utils.repr_utils import repr_dict
from arjuna.tpi.tracker import track

@track("trace")
class GuiWidgetMetaData:

    def __init__(self, locators, meta=None, process=True):
        self.__locators = None
        self.__meta = Meta(meta)
        if process:
            self.__raw_locators = locators
            self.__locators = []
            for raw_locator in self.__raw_locators:
                self.locators.append(LocatorTranslator.translate(raw_locator))
        else:
            self.__locators = locators

    @property
    def meta(self):
        return self.__meta

    @property
    def max_wait(self):
        return self.meta.max_wait

    def __str__(self):
        return repr_dict({
            "locators" : [str(l) for l in self.__locators],
            "meta" : str(self.__meta)
        })

    def print_locators(self):
        print(str(self))

    @property
    def locators(self):
        return self.__locators

    @property
    def raw_locators(self):
        return self.__raw_locators

    def create_formatted_wmd(self, **fargs):
        formatted_locators = []
        for locator in self.locators:
            formatted_locator = locator.create_formatted_locator(**fargs)
            formatted_locators.append(formatted_locator)
        return GuiWidgetMetaData(formatted_locators, meta=self.meta, process=False)

    @classmethod
    def __process_single_raw_locator(cls, impl_locator):
        ltype = impl_locator.wtype
        lvalue = impl_locator.wvalue
        p_locator = Locator(ltype=ltype, lvalue=lvalue)
        return p_locator

    @classmethod
    def convert_to_impl_with_locators(cls, *with_locators):
        out_list = []
        for locator in with_locators:
            l = isinstance(locator, ImplWith) and locator or locator.as_impl_locator()
            out_list.append(l)
        return out_list

    @classmethod
    def create_wmd(cls, *locators, meta=None):
        impl_locators = cls.convert_to_impl_with_locators(*locators)
        processed_locators = []
        for locator in impl_locators:
            ltype = locator.wtype.lower()
            if ltype == "content_locator":
                CONTENT_LOCATOR = cls.create_wmd(locator.wvalue)
                p_locator = Locator(ltype=locator.wtype, lvalue=CONTENT_LOCATOR)
            else:
                p_locator = cls.__process_single_raw_locator(locator)
            processed_locators.append(p_locator)
        return GuiWidgetMetaData(processed_locators, meta)

    @staticmethod
    def locators_as_str(locators):
        if not locators:
            return list()

        from arjuna.interact.gui.auto.finder._with import With
        out_list = []
        for l in locators:
            if isinstance(l, GuiGenericLocator) or isinstance(l, With):
                out_list.append(l.as_map())
            else:
                out_list.append(str(l))
        return out_list

class SimpleGuiWidgetMetaData(GuiWidgetMetaData):

    def __init__(self, locator_type, locator_value=dict()):
        super().__init__([Locator(ltype=locator_type, lvalue=locator_value)])



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.tracker import track

from .base import SingleGuiEntitySource

@track("trace")
class GuiElementSource(SingleGuiEntitySource):
    '''
        Abstract Base class for GUI Source of a Singular GUI entity (**GuiPage** or **GuiElement**).

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_element.source

        Args:
            raw_source: Raw XML source.
    '''

    def __init__(self, raw_source):
        super().__init__(raw_source, partial=True)
        self.__tag = None
        self.__attributes = None

    @property
    def tag(self) -> str:
        '''
            Tag name.
        '''
        return self.__tag

    def _process_elem_node(self, elem_node):
        self.__tag = elem_node.tag
        self.__attributes = elem_node.attrs

    @property
    def attrs(self) -> dict:
        '''
            All attributes as a dictionary.
        '''
        return self.__attributes

    def is_attr_present(self, attr) -> bool:
        '''
            Check if an attribute is present.

            Args:
                attr: Attribute name
        '''
        return attr in self.attrs

    def get_attr_value(self, attr, optional=False) -> str:
        '''
            Get value of an attribute.

            Args:
                attr: Attribute name

            Keyword Args:
                optional: Set to True if this attribute is optional. Default is False.

            Raises:
                Exception if optional is False and attribute is not found.
        '''
        try:
            return self.attrs[attr]
        except Exception as e:
            if optional:
                return None
            else:
                raise Exception("Attribute {} not found for element".format(attr))

    def get_value(self, optional=False):
        '''
            Get content of value attribute.

            Keyword Args:
                optional: Set to True if value attribute is optional. Default is False.

            Raises:
                Exception if optional is False and value attribute is not found.
        '''
        return self.get_attr_value("value")

    @property
    def value(self) -> str:
        '''
            Content of value attribute.
        '''
        return self.get_value()



# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from arjuna.engine.track import *

def track(level: str="debug"):
    '''
        Decorator to track a callable.

        You can track classes, methods, functions and properties using this decorator.

        Arjuna tracks the following and logs it as per **log_*** function chosen as per provided level. Default is debug.
            - Begin of call and arguments passed to the call.
            - End of the call with return value
            - Exception raised if any

        Keywrod Arguments:
            callable: Callable that is decorated.
            level: (Optional) A string representing a Logging Level (trace/debug/info/warning/error/fatal). Default is debug.

        Note:
            When you decorate a class with @track, it automatically tracks its construction, all its methods including classmethods and staticmethods.

            Properties are not auto-tracked for a class.

            You can add @track decorator to a property. If you have a property setter, adding @track to it automatically also tracks the getter.

            You can selectively add decorators to methods in the class as well.

            @track also works for functions.
    '''

    kallable = None
    kallable_type = None
    non_arg_track = False

    if inspect.isclass(level) or inspect.isfunction(level) or isinstance(level, classmethod) or isinstance(level, staticmethod) or isinstance(level, property):
        kallable = level
        non_arg_track = True
        level = "debug"
        if inspect.isclass(kallable):
            return track_class(kallable, level)
        elif inspect.isfunction(kallable):
            return track_func(level)(kallable)
        elif isinstance(kallable, classmethod):
            return classmethod(track_func(level)(kallable.__func__))
        elif isinstance(kallable, staticmethod):
            return staticmethod(track_func(level, static=True)(kallable.__func__))
        elif isinstance(kallable, property):
            return property(
                kallable.fget and track_func(level, prop=True, prop_type="fget")(kallable.fget) or None,
                kallable.fset and track_func(level, prop=True, prop_type="fset")(kallable.fset) or None,
                kallable.fdel and track_func(level, prop=True, prop_type="fdel")(kallable.fdel) or None,
            )
        else:
            raise Exception("track decorator is meant for class/method/function only.")    
    else:
        level = level is not None and level or "debug"

    def deco(kallable):
        if inspect.isclass(kallable):
            return track_class(kallable, level)
        elif inspect.isfunction(kallable) or inspect.ismethod(kallable):
            return track_func(level)(kallable)
        elif isinstance(kallable, classmethod):
            return classmethod(track_func(level)(kallable.__func__))
        elif isinstance(kallable, staticmethod):
            return staticmethod(track_func(level, static=True)(kallable.__func__))
        elif isinstance(kallable, property):
            return property(
                kallable.fget and track_func(level, prop=True, prop_type="fget")(kallable.fget) or None,
                kallable.fset and track_func(level, prop=True, prop_type="fset")(kallable.fset) or None,
                kallable.fdel and track_func(level, prop=True, prop_type="fdel")(kallable.fdel) or None,
            )
    return deco  

'''
    def __load_console(self, dl, logger):
        class __console:
            lock = threading.RLock()

            def __init__(self):
                # self.lock = threading.RLock()
                self.separator = os.linesep
                self.log_display_level = dl
                self.logger = logger

            @sync_method('lock')
            def __log(self, message, err=False):
                if err:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.error(mpart)
                else:
                    mparts = message.replace(u"\r\n", u"--|--").replace(u"\n", u"--|--").split(u'--|--')
                    for mpart in mparts:
                        self.logger.info(mpart)

            @sync_method('lock')
            def __print(self, message):
                print(message, end='', file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprint(self, message):
                print(message, end='', file=sys.stderr, flush=True)

            @sync_method('lock')
            def __println(self, message):
                print(message, file=sys.stdout, flush=True)

            @sync_method('lock')
            def __eprintln(self, message):
                print(message, file=sys.stderr, flush=True)

            @sync_method('lock')
            def __msg(self, *messages):
                return " ".join([str(m) for m in messages])

            @sync_method('lock')
            def display(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__println(message)

            @sync_method('lock')
            def display_error(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message, err=True)
                if should_print:
                    self.__eprintln(message)

            @sync_method('lock')
            def error_for_console(self, *messages):
                message = self.__msg(*messages)
                self.__eprintln(message)

            @sync_method('lock')
            def display_on_same_line(self, *messages):
                message = self.__msg(*messages)
                should_print = self.__log(message)
                if should_print:
                    self.__print(message)

            @sync_method('lock')
            def marker(self, length, symbol='-'):
                self.display(symbol * length)

            @sync_method('lock')
            def marker_error(self, length, symbol='-'):
                self.display_error(symbol * length)

            @sync_method('lock')
            def marker_on_same_line(self, length=40):
                self.display_on_same_line("-" * length)

            @sync_method('lock')
            def display_key_value(self, key, value):
                message = "%s %s".format(key, value)
                self.display(message)

            @sync_method('lock')
            def __get_formatted_key_value(self, key, value, left_padding):
                if not left_padding:
                    message = "| {:20s}| {}".format(key, value)
                else:
                    message = "| {}| {}".format(key.ljust(left_padding), value)
                return message

            @sync_method('lock')
            def __display_paddedKV(self, key, value, print_func, left_padding):
                print_func(self.__get_formatted_key_value(key, value, left_padding))

            @sync_method('lock')
            def display_padded_key_value(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display, left_padding)

            @sync_method('lock')
            def display_padded_key_value_error(self, key, value, left_padding=None):
                self.__display_paddedKV(key, value, self.display_error, left_padding)

            @sync_method('lock')
            def display_exception_block(self, e, strace):
                self.marker_error(80)
                self.display_padded_key_value_error("Exception Type", e.__class__.__name__, 30)
                self.display_padded_key_value_error("Exception Message", str(e), 30)
                self.display_padded_key_value_exception_trace("Exception Trace", strace, 30)
                self.marker_error(80)

            def set_central_log_level(self, level):
                self.central_log_level = level

            @sync_method('lock')
            def display_multiline_key_value(self, key, value, left_padding=30):
                value = str(value)
                ctrace_parts = value.replace("\t", " ").replace("\r\n\r\n","\r\n").replace("\n\n","\n").split(sys_utils.get_line_separator())

                header = self.__get_formatted_key_value(key, ctrace_parts[0], left_padding)
                self.__log(header)
                # if should_print:
                #     self.error_for_console(header)

                for s in ctrace_parts[1:]:
                    message = self.__get_formatted_key_value("", s, left_padding)
                    self.__log(message)
                    # if should_print:
                    #     self.(message)

            display_padded_key_value_exception_trace = display_multiline_key_value
        self.__console = __console()
'''